from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .exceptions import *
from .parsing import *
from ..tradedb import TradeDB, System, Station

import math

######################################################################
# Parser config

help='Find places to sell a given item within range of a given station.'
name='sell'
epilog=None
wantsTradeDB=True
arguments = [
    ParseArgument('item', help='Name of item you want to sell.', type=str),
]
switches = [
    ParseArgument(
        '--demand', '--quantity',
        help='Limit to stations known to have at least this much demand.',
        default=0,
        type=int,
    ),
    ParseArgument('--near',
            help='Find buyers within jump range of this system.',
            type=str
    ),
    ParseArgument('--ly-per',
            help='Maximum light years per jump.',
            default=None,
            dest='maxLyPer',
            metavar='N.NN',
            type=float,
    ),
    AvoidPlacesArgument(),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    BlackMarketSwitch(),
    ParseArgument('--limit',
            help='Maximum number of results to list.',
            default=None,
            type=int,
    ),
    ParseArgument('--price-sort', '-P',
            help='(When using --near) Sort by price not distance',
            action='store_true',
            default=False,
            dest='sortByPrice',
    ),
    ParseArgument('--gt',
            help='Limit to prices above Ncr',
            metavar='N',
            dest='gt',
            type="credits",
    ),
    ParseArgument('--lt',
            help='Limit to prices below Ncr',
            metavar='N',
            dest='lt',
            type="credits",
    ),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    from .commandenv import ResultRow
    
    if cmdenv.lt and cmdenv.gt:
        if cmdenv.lt <= cmdenv.gt:
            raise CommandLineError("--gt must be lower than --lt")
    
    item = tdb.lookupItem(cmdenv.item)
    cmdenv.DEBUG0("Looking up item {} (#{})", item.name(), item.ID)
    
    avoidSystems = {s for s in cmdenv.avoidPlaces if isinstance(s, System)}
    avoidStations = {s for s in cmdenv.avoidPlaces if isinstance(s, Station)}
    
    results.summary = ResultRow()
    results.summary.item = item
    results.summary.avoidSystems = avoidSystems
    results.summary.avoidStations = avoidStations
    
    if cmdenv.detail:
        avgPrice = tdb.query("""
            SELECT AVG(si.demand_price)
              FROM StationItem AS si
             WHERE si.item_id = ? AND si.demand_price > 0
        """, [item.ID]).fetchone()[0]
        results.summary.avg = int(avgPrice)
    
    # Constraints
    tables = "StationItem AS si"
    constraints = [
        "(item_id = {} AND demand_price > 0)".format(item.ID),
    ]
    columns = [
        'si.station_id',
        'si.demand_price',
        'si.demand_units',
    ]
    bindValues = []
    
    if cmdenv.demand:
        constraints.append("(demand_units >= ?)")
        bindValues.append(cmdenv.demand)
    
    if cmdenv.lt:
        constraints.append("(demand_price < ?)")
        bindValues.append(cmdenv.lt)
    if cmdenv.gt:
        constraints.append("(demand_price > ?)")
        bindValues.append(cmdenv.gt)
    
    nearSystem = cmdenv.nearSystem
    if nearSystem:
        maxLy = cmdenv.maxLyPer or tdb.maxSystemLinkLy
        results.summary.near = nearSystem
        results.summary.ly = maxLy
        distanceFn = nearSystem.distanceTo
    else:
        distanceFn = None
    
    whereClause = ' AND '.join(constraints)
    stmt = """SELECT DISTINCT {columns} FROM {tables} WHERE {where}""".format(
        columns=','.join(columns),
        tables=tables,
        where=whereClause
    )
    cmdenv.DEBUG0('SQL: {}', stmt)
    cur = tdb.query(stmt, bindValues)
    
    stationByID = tdb.stationByID
    padSize = cmdenv.padSize
    planetary = cmdenv.planetary
    fleet = cmdenv.fleet
    wantNoPlanet = cmdenv.noPlanet
    wantBlackMarket = cmdenv.blackMarket
    
    for (stationID, priceCr, demand) in cur:
        station = stationByID[stationID]
        if padSize and not station.checkPadSize(padSize):
            continue
        if fleet and not station.checkFleet(fleet):
            continue
        if planetary and not station.checkPlanetary(planetary):
            continue
        if wantNoPlanet and station.planetary != 'N':
            continue
        if wantBlackMarket and station.blackMarket != 'Y':
            continue
        if station in avoidStations:
            continue
        if station.system in avoidSystems:
            continue
        
        row = ResultRow()
        row.station = station
        if distanceFn:
            distance = distanceFn(row.station.system)
            if distance > maxLy:
                continue
            row.dist = distance
        row.price = priceCr
        row.demand = demand
        row.age = station.itemDataAgeStr
        results.rows.append(row)
    
    if not results.rows:
        raise NoDataError("No available items found")
    
    results.summary.sort = "Price"
    results.rows.sort(key=lambda result: result.demand, reverse=True)
    results.rows.sort(key=lambda result: result.price, reverse=True)
    if nearSystem and not cmdenv.sortByPrice:
        results.summary.sort = "Dist"
        results.rows.sort(key=lambda result: result.dist)
    
    limit = cmdenv.limit or 0
    if limit > 0:
        results.rows = results.rows[:limit]
    
    return results


#######################################################################
## Transform result set into output

def render(results, cmdenv, tdb):
    from ..formatting import RowFormat, ColumnFormat
    
    longestNamed = max(results.rows, key=lambda result: len(result.station.name()))
    longestNameLen = len(longestNamed.station.name())
    
    stnRowFmt = RowFormat()
    stnRowFmt.addColumn('Station', '<', longestNameLen,
            key=lambda row: row.station.name())
    stnRowFmt.addColumn('Cost', '>', 10, 'n',
            key=lambda row: row.price)
    if cmdenv.detail:
        stnRowFmt.addColumn('Demand', '>', 10,
                key=lambda row: '{:n}'.format(row.demand) if row.demand >= 0 else '?')
    if cmdenv.nearSystem:
        stnRowFmt.addColumn('DistLy', '>', 6, '.2f',
                key=lambda row: row.dist)
    
    stnRowFmt.addColumn('Age/days', '>', 7,
            key=lambda row: row.age)
    stnRowFmt.addColumn('StnLs', '>', 10,
                key=lambda row: row.station.distFromStar())
    stnRowFmt.addColumn('B/mkt', '>', 4,
            key=lambda row: TradeDB.marketStates[row.station.blackMarket])
    stnRowFmt.addColumn("Pad", '>', '3',
            key=lambda row: TradeDB.padSizes[row.station.maxPadSize])
    stnRowFmt.addColumn("Plt", '>', '3',
            key=lambda row: TradeDB.planetStates[row.station.planetary])
    stnRowFmt.addColumn("Flc", '>', '3',
            key=lambda row: TradeDB.fleetStates[row.station.fleet])
    
    if not cmdenv.quiet:
        heading, underline = stnRowFmt.heading()
        print(heading, underline, sep='\n')
    
    for row in results.rows:
        print(stnRowFmt.format(row))
    
    if cmdenv.detail:
        print("{:{lnl}} {:>10n}".format(
                "-- Average",
                results.summary.avg,
                lnl=longestNameLen,
        ))



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .parsing import *
from ..tradeexcept import TradeException
from .exceptions import CommandLineError
from ..tradedb import System, Station
from .. import prices, cache
import subprocess
import os
import sys
import pathlib

######################################################################
# Parser config

help = 'Update prices for a station.'
name = 'update'
epilog = ("Generates a human-readable version of the price list for a given station "
            "and opens it in the specified text editor.\n"
            "The format is intended to closely resemble the presentation of the "
            "market in-game. If you change the order items are listed in, "
            "the order will be kept for future edits, making it easier to quickly "
            "check for changes.")
wantsTradeDB = True
arguments = [
    ParseArgument('starting', help = 'Name of the station to update.', type = str)
]
switches = [
    ParseArgument('--timestamps', '-T',
            help = '[Text editing] Includes timestamps in the update.',
            action = 'store_true',
            default = False,
    ),
    ParseArgument('--all', '-A',
            help = 'List all known items.',
            action = 'store_true',
            default = False,
    ),
    ParseArgument('--use-demand', '-D',
            help = 'Unlock the "Demand" column in the GUI.',
            action = 'store_true',
            dest = 'useDemand',
            default = False,
    ),
    ParseArgument('--force-na', '-0',
            help = "Forces 'unk' supply to become 'n/a' by default",
            action = 'store_true',
            default = False,
            dest = 'forceNa',
    ),
    ParseArgument('--height', '-H',
            help = "[GUI] Specify height of the window",
            type = int,
            default = 800,
    ),
    ParseArgument('--front', '-F',
            help = (
                "[GUI] Keep the GUI infront of other windows; "
                "this allows you to put the window infront of "
                "the game UI if you run the game in windowed mode."
                ),
            action = 'store_true',
            default = False,
            dest = 'alwaysOnTop',
    ),
    ParseArgument('--window-x', '-wx',
            help = (
                "[GUI] Specify the window X position "
                "Use a negative value for a right-of-screen relative offset"
                ),
            default = -1,
            dest = 'windowX',
            type = int,
    ),
    ParseArgument('--window-y', '-wy',
            help = (
                "[GUI] Specify the window U position "
                "Use a negative value for a bottom-of-screen relative offset"
                ),
            default = 1,
            dest = 'windowY',
            type = int,
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--gui', '-G',
                help = "Use the experimental built-in GUI",
                action = 'store_true',
                default = False,
                dest = 'gui',
        ),
        ParseArgument('--editor',
                help = 'Generates a text file containing the prices for '
                        'the station and loads it into the specified '
                        'text editor for you.',
                default = None,
                type = str,
        ),
        ParseArgument('--sublime',
                help = 'Like --editor but uses Sublime Text (2 or 3).',
                action = 'store_const', const = 'sublime', dest = 'editing',
        ),
        ParseArgument('--notepad',
                help = 'Like --editor but uses Notepad.',
                action = 'store_const', const = 'notepad', dest = 'editing',
        ),
        ParseArgument('--npp',
                help = 'Like --editor but uses Notepad++.',
                action = 'store_const', const = 'npp', dest = 'editing',
        ),
        ParseArgument('--vim',
                help = 'Like --editor but uses vim.',
                action = 'store_const', const = 'vim', dest = 'editing',
        ),
    )
]

######################################################################
# Helpers


class TemporaryFileExistsError(TradeException):
    pass


def getTemporaryPath(cmdenv):
    tmpPath = pathlib.Path("prices.tmp")
    if tmpPath.exists():
        if not cmdenv.force:
            raise TemporaryFileExistsError(
                    "Temporary file already exists: {}\n"
                    "(Check you aren't already editing in another window"
                        .format(tmpPath)
            )
        tmpPath.unlink()
    return tmpPath


def saveTemporaryFile(tmpPath):
    if tmpPath.exists():
        lastPath = pathlib.Path("prices.last")
        if lastPath.exists():
            lastPath.unlink()
        tmpPath.rename(lastPath)


def saveCopyOfChanges(cmdenv, dbFilename, stationID):
    if "APPEND_PRICES" in os.environ:
        mode = "a"
    else:
        mode = "w"
    dumpPath = pathlib.Path("updated.prices")
    with dumpPath.open(mode, encoding = 'utf-8') as dumpFile:
        # Remember the filename so we know we need to delete it.
        prices.dumpPrices(dbFilename,
                prices.Element.full | prices.Element.blanks,
                file = dumpFile,
                stationID = stationID,
                defaultZero = False,
                debug = cmdenv.debug,
        )
    if not cmdenv.quiet:
        print("- Copy of changes saved as '{}'".format(
                str(dumpPath)
        ))


def getEditorPaths(cmdenv, editorName, envVar, windowsFolders, winExe, nixExe):
    cmdenv.DEBUG0("Locating {} editor", editorName)
    try:
        return os.environ[envVar]
    except KeyError: pass
    
    paths = []
    
    import platform
    system = platform.system()
    if system == 'Windows':
        binary = winExe
        for folder in ["Program Files", "Program Files (x86)"]:
            for version in windowsFolders:
                paths.append("{}\\{}\\{}".format(os.environ['SystemDrive'], folder, version))
    else:
        binary = nixExe
    
    try:
        paths += os.environ['PATH'].split(os.pathsep)
    except KeyError: pass
    
    for path in paths:
        candidate = os.path.join(path, binary)
        try:
            if pathlib.Path(candidate).exists():
                return candidate
        except OSError:
            pass
    
    raise CommandLineError(
        "ERROR: Unable to locate {} editor.\n"
        "Either specify the path to your editor with --editor "
        "or set the {} environment variable to point to it."
            .format(editorName, envVar)
    )


def editUpdate(tdb, cmdenv, stationID):
    """
        Dump the price data for a specific station to a file and
        launch the user's text editor to let them make changes
        to the file.
        
        If the user makes changes, re-load the file, update the
        database and regenerate the master .prices file.
    """
    
    cmdenv.DEBUG0("'update' mode with editor. editor:{} station:{}",
                    cmdenv.editor, cmdenv.origin)
    
    editor, editorArgs = cmdenv.editor, []
    if cmdenv.editing == 'sublime':
        cmdenv.DEBUG0("Sublime mode")
        editor = editor or \
                getEditorPaths(cmdenv,
                        "sublime", "SUBLIME_EDITOR",
                        ["Sublime Text 3", "Sublime Text 2"],
                        "sublime_text.exe",
                        "subl"
                )
        editorArgs += [ "--wait" ]
    elif cmdenv.editing == 'npp':
        cmdenv.DEBUG0("Notepad++ mode")
        editor = editor or \
                getEditorPaths(cmdenv,
                        "notepad++", "NOTEPADPP_EDITOR",
                        ["Notepad++"],
                        "notepad++.exe",
                        "notepad++"
                )
        if not cmdenv.quiet:
            print("NOTE: "
                    "You'll need to exit Notepad++ when you are done.")
    elif cmdenv.editing == "vim":
        cmdenv.DEBUG0("VI iMproved mode")
        if not editor:
            # Hack... Hackity hack, hack, hack.
            # This has a disadvantage in that: we don't check for just "vim" (no .exe) under Windows
            vimDirs = [
                    "Git\\share\\vim\\vim{}".format(vimVer)
                    for vimVer in range(70, 75)
            ]
            editor = getEditorPaths(cmdenv,
                    "vim",
                    "EDITOR",
                    vimDirs,
                    "vim.exe",
                    "vim"
            )
    elif cmdenv.editing == "notepad":
        cmdenv.DEBUG0("Notepad mode")
        editor = editor or "notepad.exe"  # herp
    
    try:
        envArgs = os.environ["EDITOR_ARGS"]
        if envArgs:
            editorArgs += envArgs.split(' ')
    except KeyError:
        pass
    
    # Create a temporary text file with a list of the price data.
    tmpPath = getTemporaryPath(cmdenv)
    
    absoluteFilename = None
    dbFilename = tdb.dbFilename
    try:
        elementMask = prices.Element.basic | prices.Element.supply
        if cmdenv.timestamps: elementMask |= prices.Element.timestamp
        if cmdenv.all: elementMask |= prices.Element.blanks
        # Open the file and dump data to it.
        with tmpPath.open("w", encoding = 'utf-8') as tmpFile:
            # Remember the filename so we know we need to delete it.
            absoluteFilename = str(tmpPath.resolve())
            prices.dumpPrices(dbFilename, elementMask,
                    file = tmpFile,
                    stationID = stationID,
                    defaultZero = cmdenv.forceNa,
                    debug = cmdenv.debug
            )
        
        # Stat the file so we can determine if the user writes to it.
        # Use the most recent create/modified timestamp.
        preStat = tmpPath.stat()
        preStamp = max(preStat.st_mtime, preStat.st_ctime)
        
        # Launch the editor
        editorCommandLine = [ editor ] + editorArgs + [ absoluteFilename ]
        cmdenv.DEBUG0("Invoking [{}]", ' '.join(editorCommandLine))
        try:
            result = subprocess.call(editorCommandLine)
        except FileNotFoundError:
            raise CommandLineError(
                        "Unable to launch requested editor: {}"
                            .format(editorCommandLine)
            )
        if result != 0:
            raise CommandLineError(
                    "NO DATA IMPORTED: "
                    "Your editor exited with a 'failed' exit code ({})"
                        .format(result)
            )
        
        # Did they update the file? Some editors destroy the file and rewrite it,
        # other files just write back to it, and some OSes do weird things with
        # these timestamps. That's why we have to use both mtime and ctime.
        postStat = tmpPath.stat()
        postStamp = max(postStat.st_mtime, postStat.st_ctime)
        
        if postStamp == preStamp:
            import random
            print("- No changes detected - doing nothing. {}".format(
                    random.choice([
                        "Brilliant!",
                        "I'll get my coat.",
                        "I ain't seen you.",
                        "You ain't seen me",
                        "... which was nice",
                        "Bingo!",
                        "Scorchio!",
                        "Boutros, boutros, ghali!",
                        "I'm Ed Winchester!",
                        "Suit you, sir! Oh!"
                    ])
            ))
        else:
            cache.importDataFromFile(tdb, cmdenv, tmpPath)
            saveCopyOfChanges(cmdenv, dbFilename, stationID)
        
        tmpPath.unlink()
        tmpPath = None
    
    except Exception as e:
        print("ERROR:", e)
        print()
        print("*** YOUR UPDATES WILL BE SAVED AS {} ***".format(
                "prices.last"
        ))
        # Save a copy
        if absoluteFilename and tmpPath:
            saveTemporaryFile(tmpPath)
        if "EXCEPTIONS" in os.environ:
            raise e


def guidedUpdate(tdb, cmdenv):
    dbFilename = tdb.dbFilename
    stationID = cmdenv.startStation.ID
    tmpPath = getTemporaryPath(cmdenv)
    
    cur = tdb.query("""
        SELECT  JULIANDAY('now') - JULIANDAY(MIN(modified)),
                JULIANDAY('now') - JULIANDAY(MAX(modified))
          FROM  StationItem
         WHERE  station_id = ?
    """, [stationID])
    oldest, newest = cur.fetchone()
    if oldest and newest:
        cmdenv.NOTE(
            "Current data {:.2f}-{:.2f} days old.",
            oldest, newest,
        )
    
    from .update_gui import render
    try:
        render(tdb, cmdenv, tmpPath)
        cmdenv.DEBUG0("Got results, importing")
        cache.importDataFromFile(tdb, cmdenv, tmpPath)
        
        saveCopyOfChanges(cmdenv, dbFilename, stationID)
        
        tmpPath.unlink()
        tmpPath = None
    
    except Exception as e:
        print("ERROR:", e)
        print()
        print("*** YOUR UPDATES WILL BE SAVED AS {} ***".format(
                "prices.last"
        ))
        
        if tmpPath:
            saveTemporaryFile(tmpPath)
        if "EXCEPTIONS" in os.environ:
            raise e

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    place = cmdenv.origPlace
    if isinstance(place, System):
        system = place
        if len(system.stations) != 1:
            raise CommandLineError(
                    "'{}' is a system, not a station.".format(
                        system.name()
            ))
        cmdenv.startStation = system.stations[0]
    else:
        cmdenv.startStation = place
    
    if cmdenv.gui or (not cmdenv.editor and not cmdenv.editing):
        if not cmdenv.quiet:
            print(
                    "NOTE:\n"
                    ". Press CTRL-C here to abort edit, otherwise "
                    "just close the update window to save.\n"
                    ". '-F' makes the window show in-front of the "
                    "E:D Window.\n"
                    ". '-A' forces all items to be listed.\n",
                    file = sys.stderr
            )
        guidedUpdate(tdb, cmdenv)
    else:
        # User specified one of the options to use an editor.
        editUpdate(tdb, cmdenv, cmdenv.startStation.ID)
    
    return None



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .commandenv import ResultRow
from .exceptions import *
from .parsing import *
from ..formatting import RowFormat, ColumnFormat, max_len
from ..tradedb import TradeDB

import math

######################################################################
# Parser config

# Displayed on the "trade.py --help" command list, etc.
help='Find rares near your current local.'
# Should match the name of the module minus the _cmd.py
name='rares'
# Displayed at the end of the "trade.py rares --help"
epilog=None
# Set to False in commands that need to operate without
# a trade database.
wantsTradeDB=True
# Required parameters
arguments = [
    ParseArgument(
            'near',
            help='Your current system.',
            type=str,
            metavar='SYSTEMNAME',
    ),
]
# Optional parameters
switches = [
    ParseArgument('--ly',
            help='Maximum distance to search.',
            metavar='LY',
            type=float,
            default=180,
            dest='maxLyPer',
    ),
    ParseArgument('--limit',
            help='Maximum number of results to list.',
            default=None,
            type=int,
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    ParseArgument('--price-sort', '-P',
            help='Sort by price not distance.',
            action='store_true',
            default=False,
            dest='sortByPrice',
    ),
    ParseArgument('--reverse', '-r',
            help='Reverse the list.',
            action='store_true',
            default=False,
    ),
    ParseArgument('--away',
            help='Require "--from" systems to be at least this far ' \
                    'from primary system',
            metavar='LY',
            default=0,
            type=float,
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--legal',
            help='List only items known to be legal.',
            action='store_true',
        ),
        ParseArgument('--illegal',
            help='List only items known to be illegal.',
            action='store_true',
        )
    ),
    ParseArgument('--from',
            help='Additional systems to range check candidates against, ' \
                    'requires --away.',
            metavar='SYSTEMNAME',
            action='append',
            dest='awayFrom',
    ),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    """
    Fetch all the data needed to display the results of a "rares"
    command. Does not actually print anything.
    
    Command execution is broken into two steps:
        1. cmd.run(results, cmdenv, tdb)
            Gather all the data required but generate no output,
        2. cmd.render(results, cmdenv, tdb)
            Print output to the user.
    
    This separation of concerns allows modularity; you can write
    a command that calls another command to fetch data for you
    and knowing it doesn't generate any output. Then you can
    process the data and return it and let the command parser
    decide when to turn it into output.
    
    It also opens a future door to commands that can present
    their data in a GUI as well as the command line by having
    a custom render() function.
    
    Parameters:
        results
            An object to be populated and returned
        cmdenv
            A CommandEnv object populated with the parameters
            for the command.
        tdb
            A TradeDB object to query against.
    
    Returns:
        None
            End execution without any output
        results
            Proceed to "render" with the output.
    """
    
    # Lookup the system we're currently in.
    start = cmdenv.nearSystem
    # Hoist the padSize, noPlanet and planetary parameter for convenience
    padSize = cmdenv.padSize
    noPlanet = cmdenv.noPlanet
    planetary = cmdenv.planetary
    fleet = cmdenv.fleet
    # How far we're want to cast our net.
    maxLy = float(cmdenv.maxLyPer or 0.)
    
    if cmdenv.illegal:
        wantIllegality = 'Y'
    elif cmdenv.legal:
        wantIllegality = 'N'
    else:
        wantIllegality = 'YN?'
    
    awaySystems = set()
    if cmdenv.away or cmdenv.awayFrom:
        if not cmdenv.away or not cmdenv.awayFrom:
            raise CommandLineError(
                "Invalid --away/--from usage. See --help"
            )
        minAwayDist = cmdenv.away
        for sysName in cmdenv.awayFrom:
            system = tdb.lookupPlace(sysName).system
            awaySystems.add(system)
    
    # Start to build up the results data.
    results.summary = ResultRow()
    results.summary.near = start
    results.summary.ly = maxLy
    results.summary.awaySystems = awaySystems
    
    distCheckFn = start.distanceTo
    
    # Look through the rares list.
    for rare in tdb.rareItemByID.values():
        if not rare.illegal in wantIllegality:
            continue
        if padSize:       # do we care about pad size?
            if not rare.station.checkPadSize(padSize):
                continue
        if fleet:         # do we care about fleet carrier?
            if not rare.station.checkFleet(fleet):
                continue
        if planetary:     # do we care about planetary?
            if not rare.station.checkPlanetary(planetary):
                continue
        if noPlanet and rare.station.planetary != 'N':
            continue
        rareSys = rare.station.system
        # Find the un-sqrt'd distance to the system.
        dist = distCheckFn(rareSys)
        if maxLy > 0. and dist > maxLy:
            continue
        
        if awaySystems:
            awayCheck = rareSys.distanceTo
            if any(awayCheck(away) < minAwayDist for away in awaySystems):
                continue
        
        # Create a row for this item
        row = ResultRow()
        row.rare = rare
        row.dist = dist
        results.rows.append(row)
    
    # Was anything matched?
    if not results:
        print("No matches found.")
        return None
    
    if cmdenv.sortByPrice:
        results.rows.sort(key=lambda row: row.dist)
        results.rows.sort(key=lambda row: row.rare.costCr, reverse=True)
    else:
        results.rows.sort(key=lambda row: row.rare.costCr, reverse=True)
        results.rows.sort(key=lambda row: row.dist)
    
    if cmdenv.reverse:
        results.rows.reverse()
    
    limit = cmdenv.limit or 0
    if limit > 0:
        results.rows = results.rows[:limit]
    
    return results

#######################################################################
## Transform result set into output

def render(results, cmdenv, tdb):
    """
    If the "run" command returned a result set and we are running
    from the command line, this function will be called to generate
    the output of the command.
    """
    
    if not results.rows:
        raise CommandLineError("No items found.")
    
    # Calculate the longest station and rareitem name in our list.
    longestStnNameLen = max_len(results.rows, key=lambda row: row.rare.station.name())
    longestRareNameLen = max_len(results.rows, key=lambda row: row.rare.name(cmdenv.detail))
    
    # Use the formatting system to describe what our
    # output rows are going to look at (see formatting.py)
    rowFmt = RowFormat()
    rowFmt.addColumn('Station', '<', longestStnNameLen,
            key=lambda row: row.rare.station.name())
    rowFmt.addColumn('Rare', '<', longestRareNameLen,
            key=lambda row: row.rare.name(cmdenv.detail))
    rowFmt.addColumn('Cost', '>', 10, 'n',
            key=lambda row: row.rare.costCr)
    rowFmt.addColumn('DistLy', '>', 6, '.2f',
            key=lambda row: row.dist)
    rowFmt.addColumn('Alloc', '>', 6, 'n',
            key=lambda row: row.rare.maxAlloc)
    rowFmt.addColumn('B/mkt', '>', 4,
            key=lambda row: TradeDB.marketStates[row.rare.illegal])
    rowFmt.addColumn("StnLs", '>', 10,
            key=lambda row: row.rare.station.distFromStar())
    rowFmt.addColumn('B/mkt', '>', 4,
            key=lambda row: TradeDB.marketStates[row.rare.station.blackMarket])
    rowFmt.addColumn("Pad", '>', '3',
            key=lambda row: TradeDB.padSizes[row.rare.station.maxPadSize])
    rowFmt.addColumn("Plt", '>', '3',
            key=lambda row: TradeDB.planetStates[row.rare.station.planetary])
    rowFmt.addColumn("Flc", '>', '3',
            key=lambda row: TradeDB.fleetStates[row.rare.station.fleet])
    
    # Print a heading summary if the user didn't use '-q'
    if not cmdenv.quiet:
        heading, underline = rowFmt.heading()
        print(heading, underline, sep='\n')
    
    # Print out our results.
    for row in results.rows:
        print(rowFmt.format(row))



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018, 2019
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Tools
"""This module contains various tools that could be used
externally and internally.
"""
from . import tradedb, utils

__all__ = ['test_derp']

def test_derp(tdb=None, tdenv=None):
    """
    Test whether the station names in a trade database are free of derp.
    
    Examples:
        import tradedb
        tdb = tradedb.TradeDB()
        test_derp(tdb)
        
        python -i cache.py
        >>> test_derp()
    """
    tdb = tdb or tradedb.TradeDB()
    tdenv = tdenv or tdb.tdenv
    matches = 0
    for stn in tdb.stationByID.values():
        m = utils.checkForOcrDerp(tdenv, stn.system.dbname, stn.dbname)
        if m:
            print("Match", m.groups(1))
            matches += 1
    if not matches:
        print("Current data is free of known derp")



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
import sys

from .exceptions import CommandLineError
from .parsing import *
from ..cache import buildCache
from ..tradedb import TradeDB

######################################################################
# Parser config

help = 'Build TradeDangerous cache file from sources'
name = 'buildcache'
epilog = (
        'TD will normally do this for you automatically whenever '
        'it detects changes to one or more source file; most end-'
        'users will never need to use this command.\n'
        'N.B. This process is destructive: '
        'any data in the .db that is not reflected in the '
        'source files will be lost.'
)
wantsTradeDB = False  # Cause we're about to frak with it.
arguments = [
]
switches = [
    ParseArgument(
        '--sql', default = None, dest = 'sqlFilename',
        help = 'Specify SQL script to execute.',
    ),
    ParseArgument(
        '--prices', default = None, dest = 'pricesFilename',
        help = 'Specify the prices file to load.',
    ),
    ParseArgument(
        '--force', '-f', default = False, action = 'store_true',
        dest = 'force',
        help = 'Overwrite existing file',
    ),
    ParseArgument(
        '--ignore-unknown', '-i',
        default = False, action = 'store_true',
        dest = 'ignoreUnknown',
        help = (
            "Data for systems, stations and items that are not "
            "recognized is reported as warning but skipped."
        ),
    ),
]

######################################################################
# Helpers

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    # Check that the file doesn't already exist.
    if not cmdenv.force:
        if tdb.dbPath.exists():
            raise CommandLineError(
                    "SQLite3 database '{}' already exists.\n"
                     "Either remove the file first or use the '-f' option."
                        .format(tdb.dbFilename))
    
    if not tdb.sqlPath.exists():
        raise CommandLineError(
                    "SQL File does not exist: {}"
                        .format(tdb.sqlFilename))
    
    buildCache(tdb, cmdenv)
    
    return None



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from ..tradeexcept import TradeException
import sys

######################################################################
# Exceptions

class UsageError(TradeException):
    def __init__(self, title, usage):
        self.title, self.usage = title, usage
    def __str__(self):
        return self.title + "\n\n" + self.usage


class CommandLineError(TradeException):
    """
        Raised when you provide invalid input on the command line.
        Attributes:
            errorstr       What to tell the user.
    """
    def __init__(self, errorStr, usage=None):
        self.errorStr, self.usage = errorStr, usage
    def __str__(self):
        if self.usage:
            return "ERROR: {}\n\n{}".format(self.errorStr, self.usage)
        else:
            return "ERROR: {}".format(self.errorStr)


class NoDataError(TradeException):
    """
        Raised when a request is made for which no data can be found.
        Attributes:
            errorStr        Describe the problem to the user.
    """
    def __init__(self, errorStr):
        self.errorStr = errorStr
    def __str__(self):
        return "Error: {}\n".format(self.errorStr) + ("""
Possible causes:
- No profitable runs or routes meet your criteria,
- Missing Systems or Stations along the route (see "local -vv"),
- Missing price data (see "market -vv" or "update -h"),

If you are not sure where to get data from, consider using a crowd-sourcing
project such as EDDBlink (https://github.com/eyeonus/EDDBlink).

For more help, see the TradeDangerous Wiki:
    https://github.com/eyeonus/Trade-Dangerous/wiki
""").format(sys.argv[0])


class PadSizeError(CommandLineError):
    """ Raised when an invalid pad-size option is given. """
    def __init__(self, value):
        super().__init__(
            "Invalid --pad-size '{}': Use a combination of one or more "
            "from 'S' for Small, 'M' for Medium, 'L' for Large or "
            "'?' for unknown, e.g. 'SML?' matches any pad size while "
            "'M?' matches medium or unknown or 'L' matches only large."
            .format(value)
        )


class PlanetaryError(CommandLineError):
    """ Raised when an invalid planetary option is given. """
    def __init__(self, value):
        super().__init__(
            "Invalid --planetary '{}': Use a combination of one or more "
            "from 'Y' for Yes, 'N' for No or '?' for unknown, "
            "e.g. 'YN?' matches any station while 'Y?' matches "
            "yes or unknown or 'N' matches only planetary stations."
            .format(value)
        )


class FleetCarrierError(CommandLineError):
    """ Raised when an invalid fleet-carrier option is given. """
    def __init__(self, value):
        super().__init__(
            "Invalid --fleet-carrier '{}': Use a combination of one or more "
            "from 'Y' for Yes, 'N' for No or '?' for unknown, "
            "e.g. 'YN?' matches any station while 'Y?' matches "
            "yes or unknown or 'N' matches only fleet-carrier stations."
            .format(value)
        )



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .commandenv import ResultRow
from .parsing import *
from ..formatting import RowFormat, ColumnFormat, max_len
from itertools import chain
from ..tradedb import TradeDB
from ..tradeexcept import TradeException

import math

######################################################################
# Parser config

name='local'
help='Calculate local systems.'
epilog="See also the 'station' sub-command."
wantsTradeDB=True
arguments = [
    ParseArgument(
            'near',
            help='Name of the system to query from.',
            type=str,
            metavar='SYSTEMNAME',
    ),
]
switches = [
    ParseArgument('--ly',
            help='Maximum light years from system.',
            dest='maxLyPer',
            metavar='N.NN',
            type=float,
            default=None,
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    ParseArgument('--stations',
            help='Limit to systems which have stations.',
            action='store_true',
    ),
    ParseArgument('--trading',
            help='Limit stations to ones with price data or flagged as having '
                 'a market.',
            action='store_true',
    ),
    BlackMarketSwitch(),
    ShipyardSwitch(),
    OutfittingSwitch(),
    RearmSwitch(),
    RefuelSwitch(),
    RepairSwitch(),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    cmdenv = results.cmdenv
    tdb = cmdenv.tdb
    srcSystem = cmdenv.nearSystem
    
    ly = cmdenv.maxLyPer
    if ly is None:
        ly = tdb.maxSystemLinkLy
    
    results.summary = ResultRow()
    results.summary.near = srcSystem
    results.summary.ly = ly
    results.summary.stations = 0
    
    distances = { srcSystem: 0.0 }
    
    # Calculate the bounding dimensions
    for destSys, dist in tdb.genSystemsInRange(srcSystem, ly):
        distances[destSys] = dist
    
    showStations = cmdenv.detail
    wantStations = cmdenv.stations
    padSize = cmdenv.padSize
    fleet = cmdenv.fleet
    planetary = cmdenv.planetary
    wantNoPlanet = cmdenv.noPlanet
    wantTrading = cmdenv.trading
    wantShipYard = cmdenv.shipyard
    wantBlackMarket = cmdenv.blackMarket
    wantOutfitting = cmdenv.outfitting
    wantRearm = cmdenv.rearm
    wantRefuel = cmdenv.refuel
    wantRepair = cmdenv.repair
    
    def station_filter(stations):
        for station in stations:
            if wantNoPlanet and station.planetary != 'N':
                continue
            if wantTrading and not station.isTrading:
                continue
            if wantBlackMarket and station.blackMarket != 'Y':
                continue
            if wantShipYard and station.shipyard != 'Y':
                continue
            if padSize and not station.checkPadSize(padSize):
                continue
            if fleet and not station.checkFleet(fleet):
                continue
            if planetary and not station.checkPlanetary(planetary):
                continue
            if wantOutfitting and station.outfitting != 'Y':
                continue
            if wantRearm and station.rearm != 'Y':
                continue
            if wantRefuel and station.refuel != 'Y':
                continue
            if wantRepair and station.repair != 'Y':
                continue
            yield station
    
    for (system, dist) in sorted(distances.items(), key=lambda x: x[1]):
        if showStations or wantStations:
            stations = []
            for (station) in station_filter(system.stations):
                stations.append(
                    ResultRow(
                        station=station,
                        age=station.itemDataAgeStr,
                    )
                )
            if not stations and wantStations:
                continue
        
        row = ResultRow()
        row.system = system
        row.dist = dist
        row.stations = stations if showStations else []
        results.rows.append(row)
        results.summary.stations += len(row.stations)
    
    return results

######################################################################
# Transform result set into output

def render(results, cmdenv, tdb):
    if not results or not results.rows:
        raise TradeException(
            "No systems found within {}ly of {}."
            .format(results.summary.ly, results.summary.near.name())
        )
    
    # Compare system names so we can tell
    maxSysLen = max_len(results.rows, key=lambda row: row.system.name())
    
    sysRowFmt = RowFormat().append(
        ColumnFormat("System", '<', maxSysLen,
                key=lambda row: row.system.name())
    ).append(
        ColumnFormat("Dist", '>', '7', '.2f',
                key=lambda row: row.dist)
    )
    
    showStations = cmdenv.detail
    if showStations:
        maxStnLen = max_len(
            chain.from_iterable(row.stations for row in results.rows),
            key=lambda row: row.station.dbname
        )
        maxLsLen = max_len(
            chain.from_iterable(row.stations for row in results.rows),
            key=lambda row: row.station.distFromStar()
        )
        maxLsLen = max(maxLsLen, 5)
        stnRowFmt = RowFormat(prefix='  /  ').append(
                ColumnFormat("Station", '.<', maxStnLen + 2,
                    key=lambda row: row.station.dbname)
        ).append(
                ColumnFormat("StnLs", '>', maxLsLen,
                    key=lambda row: row.station.distFromStar())
        ).append(
                ColumnFormat("Age/days", '>', 7,
                        key=lambda row: row.age)
        ).append(
                ColumnFormat("Mkt", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.market])
        ).append(
                ColumnFormat("BMk", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.blackMarket])
        ).append(
                ColumnFormat("Shp", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.shipyard])
        ).append(
                ColumnFormat("Out", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.outfitting])
        ).append(
                ColumnFormat("Arm", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.rearm])
        ).append(
                ColumnFormat("Ref", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.refuel])
        ).append(
                ColumnFormat("Rep", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.repair])
        ).append(
                ColumnFormat("Pad", '>', '3',
                    key=lambda row: \
                        TradeDB.padSizes[row.station.maxPadSize])
        ).append(
                ColumnFormat("Plt", '>', '3',
                    key=lambda row: \
                        TradeDB.planetStates[row.station.planetary])
        ).append(
                ColumnFormat("Flc", '>', '3',
                    key=lambda row: \
                        TradeDB.fleetStates[row.station.fleet])
        )
        if cmdenv.detail > 1:
            stnRowFmt.append(
                ColumnFormat("Itms", ">", 4,
                    key=lambda row: row.station.itemCount)
            )
    
    cmdenv.DEBUG0(
        "Systems within {ly:<5.2f}ly of {sys}.\n",
        sys=results.summary.near.name(),
        ly=results.summary.ly,
    )
    
    if not cmdenv.quiet:
        heading, underline = sysRowFmt.heading()
        if showStations:
            print(heading)
            heading, underline = stnRowFmt.heading()
        print(heading, underline, sep='\n')
    
    for row in results.rows:
        print(sysRowFmt.format(row))
        for stnRow in row.stations:
            print(stnRowFmt.format(stnRow))



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018, 2019
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Database Module

"""
Provides the primary classes used within TradeDangerous:

TradeDB, System, Station, Ship, Item, RareItem and Trade.

These classes are primarily for describing the database.

Simplistic use might be:
    
    import tradedb
    
    # Create an instance: You can specify a debug level as a
    # parameter, for more advanced configuration, see the
    # tradeenv.TradeEnv() class.
    tdb = tradedb.TradeDB()
    
    # look up a System by name
    sol = tdb.lookupSystem("SOL")
    ibootis = tdb.lookupSystem("i BootiS")
    ibootis = tdb.lookupSystem("ibootis")
    
    # look up a Station by name
    abe = tdb.lookupStation("Abraham Lincoln")
    abe = tdb.lookupStation("Abraham Lincoln", sol)
    abe = tdb.lookupStation("hamlinc")
    
    # look up something that could be a system or station,
    # where 'place' syntax can be:
    #  SYS, STN, SYS/STN, @SYS, /STN or @SYS/STN
    abe = tdb.lookupPlace("Abraham Lincoln")
    abe = tdb.lookupPlace("HamLinc")
    abe = tdb.lookupPlace("@SOL/HamLinc")
    abe = tdb.lookupPlace("so/haml")
    abe = tdb.lookupPlace("sol/abraham lincoln")
    abe = tdb.lookupPlace("@sol/abrahamlincoln")
    james = tdb.lookupPlace("shin/jamesmem")
"""

######################################################################
# Imports


from collections import namedtuple, defaultdict
from pathlib import Path
from .tradeenv import TradeEnv
from .tradeexcept import TradeException

from . import cache, fs
import heapq
import itertools
import locale
import math
import os
import re
import sqlite3
import sys

haveNumpy = False
try:
    import numpy
    import numpy.linalg
    haveNumpy = True
except (KeyError, ImportError):
    pass
if not haveNumpy:
    class numpy(object):
        array = False
        float32 = False
        ascontiguousarray = False
        class linalg(object):
            norm = False

locale.setlocale(locale.LC_ALL, '')

######################################################################
# Classes

class AmbiguityError(TradeException):
    """
        Raised when a search key could match multiple entities.
        Attributes:
            lookupType - description of what was being queried,
            searchKey  - the key given to the search routine,
            anyMatch - list of anyMatch
            key        - retrieve the display string for a candidate
    """
    def __init__(
            self, lookupType, searchKey, anyMatch, key=lambda item: item
            ):
        self.lookupType = lookupType
        self.searchKey = searchKey
        self.anyMatch = anyMatch
        self.key = key
    
    def __str__(self):
        anyMatch, key = self.anyMatch, self.key
        if len(anyMatch) > 10:
            opportunities = ", ".join([
                key(c) for c in anyMatch[:10]
            ] + ["..."])
        else:
            opportunities = ", ".join(
                key(c) for c in anyMatch[0:-1]
            )
            opportunities += " or " + key(anyMatch[-1])
        return '{} "{}" could match {}'.format(
            self.lookupType, str(self.searchKey),
            opportunities
        )

class SystemNotStationError(TradeException):
    """
        Raised when a station lookup matched a System but
        could not be automatically reduced to a Station.
    """
    pass

######################################################################


def makeStellarGridKey(x, y, z):
    """
    The Stellar Grid is a map of systems based on their Stellar
    co-ordinates rounded down to 32lys. This makes it much easier
    to find stars within rectangular volumes.
    """
    return (int(x) >> 5, int(y) >> 5, int(z) >> 5)

class System(object):
    """
    Describes a star system which may contain one or more Station objects.
    
    Caution: Do not use _rangeCache directly, use TradeDB.genSystemsInRange.
    """
    
    __slots__ = (
        'ID',
        'dbname', 'posX', 'posY', 'posZ', 'pos', 'stations',
        'addedID',
        '_rangeCache'
    )
    
    class RangeCache(object):
        """
        Lazily populated cache of neighboring systems.
        """
        def __init__(self):
            self.systems = []
            self.probedLy = 0.
    
    def __init__(
            self, ID, dbname, posX, posY, posZ, addedID,
            ary=numpy.array,
            nptype=numpy.float32,
            ):
        self.ID = ID
        self.dbname = dbname
        self.posX, self.posY, self.posZ = posX, posY, posZ
        if ary:
            self.pos = ary([posX, posY, posZ], nptype)
        self.addedID = addedID or 0
        self.stations = ()
        self._rangeCache = None
    
    @property
    def system(self):
        return self
    
    def distToSq(self, other):
        """
        Returns the square of the distance between two systems.
        
        It is slightly cheaper to calculate the square of the
        distance between two points, so when you are primarily
        doing distance checks you can use this less expensive
        distance query and only perform a sqrt (** 0.5) on the
        distances that fall within your constraint.
        
        Args:
            other:
                The other System to measure the distance between.
        
        Returns:
            Distance in light years to the power of 2 (i.e. squared).
        
        Example:
            # Calculate which of [systems] is within 12 ly
            # of System "target".
            maxLySq = 12 ** 2   # Maximum of 12 ly.
            inRange = []
            for sys in systems:
                if sys.distToSq(target) <= maxLySq:
                    inRange.append(sys)
        """
        return (
            (self.posX - other.posX) ** 2 +
            (self.posY - other.posY) ** 2 +
            (self.posZ - other.posZ) ** 2
        )
    
    def distanceTo(self, other):
        """
        Returns the distance (in ly) between two systems.
        
        NOTE: If you are primarily testing/comparing
        distances, consider using "distToSq" for the test.
        
        Returns:
            Distance in light years.
        
        Example:
            print("{} -> {}: {} ly".format(
                lhs.name(), rhs.name(),
                lhs.distanceTo(rhs),
            ))
        """
        return (
            (self.posX - other.posX) ** 2 +
            (self.posY - other.posY) ** 2 +
            (self.posZ - other.posZ) ** 2
        ) ** 0.5  # fast sqrt
    
    if haveNumpy:
        def all_distances(
                self, iterable,
                ary=numpy.ascontiguousarray, norm=numpy.linalg.norm,
                ):
            """
            Takes a list of systems and returns their distances from this system.
            """
            return numpy.linalg.norm(
                ary([s.pos for s in iterable]) - self.pos,
                ord=2, axis=1.
            )
    
    def getStation(self, stationName):
        """
        Quick case-insensitive lookup of a station name within the
        stations in this system.
        
        Returns:
            Station() object if a match is found,
            otherwise None.
        """
        upperName = stationName.upper()
        for station in self.stations:
            if station.dbname.upper() == upperName:
                return station
        return None
    
    def name(self, detail=0):
        return self.dbname
    
    def str(self):
        return self.dbname

######################################################################

class Destination(namedtuple('Destination', [
        'system', 'station', 'via', 'distLy'
        ])):
    pass

class DestinationNode(namedtuple('DestinationNode', [
        'system', 'via', 'distLy'
        ])):
    pass

class Station(object):
    """
    Describes a station (trading or otherwise) in a system.
    
    For obtaining trade information for a given station see one of:
        TradeCalc.getTrades        (fast and cheap)
    """
    __slots__ = (
        'ID', 'system', 'dbname',
        'lsFromStar', 'market', 'blackMarket', 'shipyard', 'maxPadSize',
        'outfitting', 'rearm', 'refuel', 'repair', 'planetary','fleet',
        'itemCount', 'dataAge',
    )
    
    def __init__(
            self, ID, system, dbname,
            lsFromStar, market, blackMarket, shipyard, maxPadSize,
            outfitting, rearm, refuel, repair, planetary, fleet,
            itemCount=0, dataAge=None,
            ):
        self.ID, self.system, self.dbname = ID, system, dbname
        self.lsFromStar = int(lsFromStar)
        self.market = market if itemCount == 0 else 'Y'
        self.blackMarket = blackMarket
        self.shipyard = shipyard
        self.maxPadSize = maxPadSize
        self.outfitting = outfitting
        self.rearm = rearm
        self.refuel = refuel
        self.repair = repair
        self.planetary = planetary
        self.fleet = fleet
        self.itemCount = itemCount
        self.dataAge = dataAge
        system.stations = system.stations + (self,)
    
    def name(self, detail=0):
        return '%s/%s' % (self.system.dbname, self.dbname)
    
    def checkPadSize(self, maxPadSize):
        """
        Tests if the Station's max pad size matches one of the
        values in 'maxPadSize'.
        
        Args:
            maxPadSize
                A string of one or more max pad size values that
                you want to match against.
        
        Returns:
            True
                If self.maxPadSize is None or empty, or matches a
                member of maxPadSize
            False
                If maxPadSize was not empty but self.maxPadSize
                did not match it.
        
        Examples:
            # Require a medium max pad size - not small or large
            station.checkPadSize("M")
            # Require medium or unknown
            station.checkPadSize("M?")
            # Require small, large or unknown
            station.checkPadSize("SL?")
        """
        return (not maxPadSize or self.maxPadSize in maxPadSize)
    
    def checkPlanetary(self, planetary):
        """
        Tests if the Station's planetary matches one of the
        values in 'planetary'.
        
        Args:
            planetary
                A string of one or more planetary values that
                you want to match against.
        
        Returns:
            True
                If self.planetary is None or empty, or matches a
                member of planetary
            False
                If planetary was not empty but self.planetary
                did not match it.
        
        Examples:
            # Require a planetary station
            station.checkPlanetary("Y")
            # Require planetary or unknown
            station.checkPlanetary("Y?")
            # Require no planetary station
            station.checkPlanetary("N")
        """
        return (not planetary or self.planetary in planetary)
    
    def checkFleet(self, fleet):
        """
        Same as checkPlanetary, but for fleet carriers.
        """
        return (not fleet or self.fleet in fleet)
    
    
    def distFromStar(self, addSuffix=False):
        """
        Returns a textual description of the distance from this
        Station to the parent star.
        
        Args:
            addSuffix[=False]:
                Always add a unit suffix (ls, Kls, ly)
        """
        ls = self.lsFromStar
        if not ls:
            if addSuffix:
                return "Unk"
            else:
                return '?'
        if ls < 1000:
            suffix = 'ls' if addSuffix else ''
            return '{:n}'.format(ls)+suffix
        if ls < 10000:
            suffix = 'ls' if addSuffix else ''
            return '{:.2f}K'.format(ls / 1000)+suffix
        if ls < 1000000:
            suffix = 'ls' if addSuffix else ''
            return '{:n}K'.format(int(ls / 1000))+suffix
        return '{:.2f}ly'.format(ls / (365*24*60*60))
    
    @property
    def isTrading(self):
        """
        True if the station is thought to be trading.
        
        A station is considered 'trading' if it has an item count > 0 or
        if it's "market" column is flagged 'Y'.
        """
        return (self.itemCount > 0 or self.market == 'Y')
    
    @property
    def itemDataAgeStr(self):
        """ Returns the age in days of item data if present, else "-". """
        if self.itemCount and self.dataAge:
            return "{:7.2f}".format(self.dataAge)
        return "-"
    
    def str(self):
        return '%s/%s' % (self.system.dbname, self.dbname)

######################################################################


class Ship(namedtuple('Ship', (
        'ID', 'dbname', 'cost', 'fdevID', 'stations'
        ))):
    """
    Ship description.
    
    Attributes:
        ID          -- The database ID
        dbname      -- The name as present in the database
        cost        -- How many credits to buy
        fdevID      -- FDevID as provided by the companion API.
        stations    -- List of Stations ship is sold at.
    """
    
    def name(self, detail=0):
        return self.dbname

######################################################################


class Category(namedtuple('Category', (
        'ID', 'dbname', 'items'
        ))):
    """
    Item Category
    
    Items are organized into categories (Food, Drugs, Metals, etc).
    Category object describes a category's ID, name and list of items.
    
    Attributes:
        ID
            The database ID
        dbname
            The name as present in the database.
        items
            List of Item objects within this category.
    
    Member Functions:
        name()
            Returns the display name for this Category.
    """
    
    def name(self, detail=0):
        return self.dbname.upper()

######################################################################


class Item(object):
    """
    A product that can be bought/sold in the game.
    
    Attributes:
        ID       -- Database ID.
        dbname   -- Name as it appears in-game and in the DB.
        category -- Reference to the category.
        fullname -- Combined category/dbname for lookups.
        avgPrice -- Galactic average as shown in game.
        fdevID   -- FDevID as provided by the companion API.
    """
    __slots__ = ('ID', 'dbname', 'category', 'fullname', 'avgPrice', 'fdevID')
    
    def __init__(self, ID, dbname, category, fullname, avgPrice=None, fdevID=None):
        self.ID = ID
        self.dbname = dbname
        self.category = category
        self.fullname = fullname
        self.avgPrice = avgPrice
        self.fdevID   = fdevID
    
    def name(self, detail=0):
        return self.fullname if detail > 0 else self.dbname

######################################################################


class RareItem(namedtuple('RareItem', (
        'ID', 'station', 'dbname', 'costCr', 'maxAlloc', 'illegal',
        'suppressed', 'category', 'fullname',
        ))):
    """
    Describes a RareItem from the database.
    
    Attributes:
        ID         -- Database ID,
        station    -- Which Station this is bought from,
        dbname     -- The name are presented in the database,
        costCr     -- Buying price.
        maxAlloc   -- How many the player can carry at a time,
        illegal    -- If the item may be considered illegal,
        suppressed -- The item is suppressed.
        category   -- Reference to the category.
        fullname   -- Combined category/dbname.
    """
    
    def name(self, detail=0):
        return self.fullname if detail > 0 else self.dbname

######################################################################


class Trade(namedtuple('Trade', (
        'item',
        'costCr', 'gainCr',
        'supply', 'supplyLevel',
        'demand', 'demandLevel',
        'srcAge', 'dstAge'
        ))):
    """
    Describes what it would cost and how much you would gain
    when selling an item between two specific stations.
    """
    def name(self, detail=0):
        return self.item.name(detail=detail)

######################################################################


class TradeDB(object):
    """
    Encapsulation for the database layer.
    
    Attributes:
        dataPath
            Path() to the data directory
        dbPath
            Path() of the .db location
        tradingCount
            Number of "profitable trade" items processed
        tradingStationCount
            Number of stations trade data has been loaded for
        tdenv
            The TradeEnv associated with this TradeDB
        sqlPath
            Path() of the .sql file
        pricesPath
            Path() of the .prices file
        importTables
            List of the .csv files
    
    Static methods:
        calculateDistance2(lx, ly, lz, rx, ry, rz)
            Returns the square of the distance in ly between two points.
        
        calculateDistance(lx, ly, lz, rx, ry, rz)
            Returns the distance in ly between two points.
        
        listSearch(...)
            Performs partial and ambiguity matching of a word from a list
            of potential values.
        
        normalizedStr(text)
            Case and punctuation normalizes a string to make it easier
            to find approximate matches.
    """
    
    # Translation map for normalizing strings
    normalizeTrans = str.maketrans(
        'abcdefghijklmnopqrstuvwxyz',
        'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        '[]()*+-.,{}:'
        )
    trimTrans = str.maketrans('', '', ' \'')
    
    # The DB cache
    defaultDB = 'TradeDangerous.db'
    # File containing SQL to build the DB cache from
    defaultSQL = 'TradeDangerous.sql'
    # File containing text description of prices
    defaultPrices = 'TradeDangerous.prices'
    # array containing standard tables, csvfilename and tablename
    # WARNING: order is important because of dependencies!
    defaultTables = (
        ('Added.csv', 'Added'),
        ('System.csv', 'System'),
        ('Station.csv', 'Station'),
        ('Ship.csv', 'Ship'),
        ('ShipVendor.csv', 'ShipVendor'),
        ('Upgrade.csv', 'Upgrade'),
        ('UpgradeVendor.csv', 'UpgradeVendor'),
        ('Category.csv', 'Category'),
        ('Item.csv', 'Item'),
        ('RareItem.csv', 'RareItem'),
        ('FDevShipyard.csv', 'FDevShipyard'),
        ('FDevOutfitting.csv', 'FDevOutfitting'),
    )
    
    # Translation matrixes for attributes -> common presentation
    marketStates = planetStates = fleetStates = {'?': '?', 'Y': 'Yes', 'N': 'No'}
    marketStatesExt = planetStatesExt = fleetStatesExt = {'?': 'Unk', 'Y': 'Yes', 'N': 'No'}
    padSizes = {'?': '?', 'S': 'Sml', 'M': 'Med', 'L': 'Lrg'}
    padSizesExt = {'?': 'Unk', 'S': 'Sml', 'M': 'Med', 'L': 'Lrg'}
    
    def __init__(
            self,
            tdenv=None,
            load=True,
            debug=None,
            ):
        self.conn = None
        self.cur = None
        self.tradingCount = None
        
        tdenv = tdenv or TradeEnv(debug=(debug or 0))
        self.tdenv = tdenv
        
        self.templatePath = Path(tdenv.templateDir).resolve()
        self.dataPath = dataPath = fs.ensurefolder(tdenv.dataDir)
        
        fs.copy_if_newer((self.templatePath / Path("Added.csv")), (self.dataPath / Path("Added.csv")))
        fs.copy_if_newer((self.templatePath / Path("RareItem.csv")), (self.dataPath / Path("RareItem.csv")))
        fs.copy_if_newer((self.templatePath / Path("TradeDangerous.sql")), (self.dataPath / Path("TradeDangerous.sql")))
        
        self.dbPath = Path(tdenv.dbFilename or dataPath / TradeDB.defaultDB)
        self.sqlPath = dataPath / Path(tdenv.sqlFilename or TradeDB.defaultSQL)
        pricePath = Path(tdenv.pricesFilename or TradeDB.defaultPrices)
        self.pricesPath = dataPath / pricePath
        self.importTables = [
            (str(dataPath / Path(fn)), tn)
            for fn, tn in TradeDB.defaultTables
        ]
        self.importPaths = {tn: tp for tp, tn in self.importTables}
        
        self.dbFilename = str(self.dbPath)
        self.sqlFilename = str(self.sqlPath)
        self.pricesFilename = str(self.pricesPath)
        
        self.avgSelling, self.avgBuying = None, None
        self.tradingStationCount = 0
        
        if load:
            self.reloadCache()
            self.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)
    
    @staticmethod
    def calculateDistance2(lx, ly, lz, rx, ry, rz):
        """
        Returns the distance in ly between two points.
        """
        dX = (lx - rx)
        dY = (ly - ry)
        dZ = (lz - rz)
        distSq = (dX ** 2) + (dY ** 2) + (dZ ** 2)
        return distSq
    
    @staticmethod
    def calculateDistance(lx, ly, lz, rx, ry, rz):
        """
        Returns the distance in ly between two points.
        """
        dX = (lx - rx)
        dY = (ly - ry)
        dZ = (lz - rz)
        distSq = (dX ** 2) + (dY ** 2) + (dZ ** 2)
        return distSq ** 0.5
    
    ############################################################
    # Access to the underlying database.
    
    def getDB(self):
        if self.conn:
            return self.conn
        self.tdenv.DEBUG1("Connecting to DB")
        conn = sqlite3.connect(self.dbFilename)
        conn.execute("PRAGMA foreign_keys=ON")
        conn.execute("PRAGMA synchronous=OFF")
        conn.execute("PRAGMA temp_store=MEMORY")
        conn.create_function('dist2', 6, TradeDB.calculateDistance2)
        return conn
    
    def query(self, *args):
        """ Perform an SQL query on the DB and return the cursor. """
        conn = self.getDB()
        cur = conn.cursor()
        cur.execute(*args)
        return cur
    
    def queryColumn(self, *args):
        """ perform an SQL query and return a single column. """
        return self.query(args).fetchone()[0]
    
    def reloadCache(self):
        """
        Checks if the .sql, .prices or *.csv files are newer than the cache.
        """
        
        if self.dbPath.exists():
            dbFileStamp = self.dbPath.stat().st_mtime
            
            paths = [self.sqlPath]
            paths += [Path(f) for (f, _) in self.importTables]
            
            changedPaths = [
                [path, path.stat().st_mtime]
                for path in paths
                if path.exists() and path.stat().st_mtime > dbFileStamp
            ]
            
            if not changedPaths:
                # Do we need to reload the .prices file?
                if not self.pricesPath.exists():
                    self.tdenv.DEBUG1("No .prices file to load")
                    return
                
                pricesStamp = self.pricesPath.stat().st_mtime
                if pricesStamp <= dbFileStamp:
                    self.tdenv.DEBUG1("DB Cache is up to date.")
                    return
                
                self.tdenv.DEBUG0(".prices has changed: re-importing")
                cache.importDataFromFile(
                    self, self.tdenv, self.pricesPath, reset=True
                )
                return
            
            self.tdenv.DEBUG0("Rebuilding DB Cache [{}]", str(changedPaths))
        else:
            self.tdenv.DEBUG0("Building DB Cache")
        
        cache.buildCache(self, self.tdenv)
    
    ############################################################
    # Load "added" data.
    
    def _loadAdded(self):
        """
        Loads the Added table as a simple dictionary
        """
        stmt = """
            SELECT added_id, name
              FROM Added
        """
        self.cur.execute(stmt)
        addedByID = {}
        for ID, name in self.cur:
            addedByID[ID] = name
        self.addedByID = addedByID
        self.tdenv.DEBUG1("Loaded {:n} Addeds", len(addedByID))
    
    def lookupAdded(self, name):
        name = name.lower()
        for ID, added in self.addedByID.items():
            if added.lower() == name:
                return ID
        raise KeyError(name)
    
    ############################################################
    # Star system data.
    
    def systems(self):
        """ Iterate through the list of systems. """
        yield from self.systemByID.values()
    
    def _loadSystems(self):
        """
        Initial load the (raw) list of systems.
        CAUTION: Will orphan previously loaded objects.
        """
        stmt = """
                SELECT system_id,
                       name, pos_x, pos_y, pos_z,
                       added_id
                  FROM System
            """
        self.cur.execute(stmt)
        systemByID, systemByName = {}, {}
        for (ID, name, posX, posY, posZ, addedID) in self.cur:
            system = System(ID, name, posX, posY, posZ, addedID)
            systemByID[ID] = systemByName[name.upper()] = system
        
        self.systemByID, self.systemByName = systemByID, systemByName
        self.tdenv.DEBUG1("Loaded {:n} Systems", len(systemByID))
    
    def lookupSystem(self, key):
        """
        Look up a System object by it's name.
        """
        if isinstance(key, System):
            return key
        if isinstance(key, Station):
            return key.system
        
        return TradeDB.listSearch(
            "System", key, self.systems(), key=lambda system: system.dbname
        )
    
    def addLocalSystem(
            self,
            name,
            x, y, z,
            added="Local",
            modified='now',
            commit=True,
            ):
        """
        Add a system to the local cache and memory copy.
        """
        
        db = self.getDB()
        cur = db.cursor()
        cur.execute("""
                INSERT INTO System (
                    name, pos_x, pos_y, pos_z, added_id, modified
                ) VALUES (
                    ?, ?, ?, ?,
                    (SELECT added_id FROM Added WHERE name = ?),
                    DATETIME(?)
                )
        """, [
            name, x, y, z, added, modified,
        ])
        ID = cur.lastrowid
        system = System(ID, name.upper(), x, y, z, 0)
        self.systemByID[ID] = system
        self.systemByName[system.dbname] = system
        if commit:
            db.commit()
        self.tdenv.NOTE(
            "Added new system #{}: {} [{},{},{}]",
            ID, name, x, y, z
        )
        # Invalidate the grid
        self.stellarGrid = None
        return system
    
    def updateLocalSystem(
            self, system,
            name, x, y, z, added="Local", modified='now',
            force=False,
            commit=True,
            ):
        """
        Updates an entry for a local system.
        """
        oldname = system.dbname
        dbname = name.upper()
        if not force:
            if oldname == dbname and \
                    system.posX == x and \
                    system.posY == y and \
                    system.posZ == z:
                return False
        del self.systemByName[oldname]
        db = self.getDB()
        db.execute("""
            UPDATE System
               SET name=?,
                   pos_x=?, pos_y=?, pos_z=?,
                   added_id=(SELECT added_id FROM Added WHERE name = ?),
                   modified=DATETIME(?)
             WHERE system_id = ?
        """, [
            dbname, x, y, z, added, modified,
            system.ID,
        ])
        if commit:
            db.commit()
        self.tdenv.NOTE(
            "{} (#{}) updated in {}: {}, {}, {}, {}, {}, {}",
            oldname, system.ID,
            self.dbPath if self.tdenv.detail > 1 else "local db",
            dbname,
            x, y, z,
            added, modified,
        )
        self.systemByName[dbname] = system
        
        return True
    
    def removeLocalSystem(
            self, system,
            commit=True,
        ):
        """ Removes a system and it's stations from the local DB. """
        for stn in self.stations:
            self.removeLocalStation(stn, commit=False)
        db = self.getDB()
        db.execute("""
            DELETE FROM System WHERE system_id = ?
        """, [
            system.ID
        ])
        if commit:
            db.commit()
        del self.systemByName[system.dbname]
        del self.systemByID[system.ID]
        
        self.tdenv.NOTE(
            "{} (#{}) deleted from {}",
            system.name(), system.ID,
            self.dbPath if self.tdenv.detail > 1 else "local db",
        )
        
        system.dbname = "DELETED " + system.dbname
        del system
    
    def __buildStellarGrid(self):
        """
        Divides the galaxy into a fixed-sized grid allowing us to
        aggregate small numbers of stars by locality.
        """
        stellarGrid = self.stellarGrid = dict()
        for system in self.systemByID.values():
            key = makeStellarGridKey(system.posX, system.posY, system.posZ)
            try:
                grid = stellarGrid[key]
            except KeyError:
                grid = stellarGrid[key] = []
            grid.append(system)
    
    def genStellarGrid(self, system, ly):
        """
        Yields Systems within a given radius of a specified System.
        
        Args:
            system:
                The System to center the search on,
            ly:
                The radius of the search around system,
        
        Yields:
            (candidate, distLySq)
                candidate:
                    System that was found,
                distLySq:
                    The *SQUARE* of the distance in light-years
                    between system and candidate.
        """
        if self.stellarGrid is None:
            self.__buildStellarGrid()
        
        sysX, sysY, sysZ = system.posX, system.posY, system.posZ
        lwrBound = makeStellarGridKey(sysX - ly, sysY - ly, sysZ - ly)
        uprBound = makeStellarGridKey(sysX + ly, sysY + ly, sysZ + ly)
        lySq = ly ** 2
        stellarGrid = self.stellarGrid
        for x in range(lwrBound[0], uprBound[0]+1):
            for y in range(lwrBound[1], uprBound[1]+1):
                for z in range(lwrBound[2], uprBound[2]+1):
                    try:
                        grid = stellarGrid[(x, y, z)]
                    except KeyError:
                        continue
                    for candidate in grid:
                        distSq = (candidate.posX - sysX) ** 2
                        if distSq > lySq:
                            continue
                        distSq += (candidate.posY - sysY) ** 2
                        if distSq > lySq:
                            continue
                        distSq += (candidate.posZ - sysZ) ** 2
                        if distSq > lySq:
                            continue
                        if candidate is not system:
                            yield candidate, distSq ** 0.5
    
    def genSystemsInRange(self, system, ly, includeSelf=False):
        """
        Yields Systems within a given radius of a specified System.
        Results are sorted by distance and cached for subsequent
        queries in the same run.
        
        Args:
            system:
                The System to center the search on,
            ly:
                The radius of the search around system,
            includeSelf:
                Whether to include 'system' in the results or not.
        
        Yields:
            (candidate, distLy)
                candidate:
                    System that was found,
                distLy:
                    The distance in lightyears between system and candidate.
        """
        
        cache = system._rangeCache
        if not cache:
            cache = system._rangeCache = System.RangeCache()
        cachedSystems = cache.systems
        
        if ly > cache.probedLy:
            # Consult the database for stars we haven't seen.
            cachedSystems = cache.systems = list(
                self.genStellarGrid(system, ly)
            )
            cachedSystems.sort(key=lambda ent: ent[1])
            cache.probedLy = ly
        
        if includeSelf:
            yield system, 0.
        
        if cache.probedLy > ly:
            # Cache may contain values outside our view
            for sys, dist in cachedSystems:
                if dist <= ly:
                    yield sys, dist
        else:
            # No need to be conditional inside the loop
            yield from cachedSystems
    
    def getRoute(self, origin, dest, maxJumpLy, avoiding=[], stationInterval=0):
        """
        Find a shortest route between two systems with an additional
        constraint that each system be a maximum of maxJumpLy from
        the previous system.
        
        Args:
            origin:
                System (or station) to start from,
            dest:
                System (or station) to terminate at,
            maxJumpLy:
                Maximum light years between systems,
            avoiding:
                List of systems being avoided
            stationInterval:
                If non-zero, require a station at least this many jumps,
            tdenv.padSize:
                Controls the pad size of stations for refuelling
        
        Returns:
            None
                No route was found
            
            [(origin, 0),...(dest, N)]
                A list of (system, distanceSoFar) values describing
                the route.
        
        Example:
            If there are systems A, B and C such
            that A->B is 7ly and B->C is 8ly then:
                
                origin = lookupPlace("A")
                dest = lookupPlace("C")
                route = tdb.getRoute(origin, dest, 9)
            
            The route should be:
                
                [(System(A), 0), (System(B), 7), System(C), 15)]
        
        """
        
        if isinstance(origin, Station):
            origin = origin.system
        if isinstance(dest, Station):
            dest = dest.system
        
        if origin == dest:
            return ((origin, 0), (dest, 0))
        
        # openSet is the list of nodes we want to visit, which will be
        # used as a priority queue (heapq).
        # Each element is a tuple of the 'priority' (the combination of
        # the total distance to the node and the distance left from the
        # node to the destination.
        openSet = [(0, 0, origin.ID, 0)]
        # Track predecessor nodes for everwhere we visit
        distances = {origin: (None, 0)}
        
        if avoiding:
            if dest in avoiding:
                raise ValueError("Destination is in avoidance list")
            for avoid in avoiding:
                if isinstance(avoid, System):
                    distances[avoid] = (None, -1)
        
        systemsInRange = self.genSystemsInRange
        heappop  = heapq.heappop
        heappush = heapq.heappush
        distTo = float("inf")
        defaultDist = (None, distTo)
        getDist  = distances.get
        
        destID = dest.ID
        sysByID = self.systemByID
        
        maxPadSize = self.tdenv.padSize
        if not maxPadSize:
            checkStations = lambda system: bool(system.stations())
        else:
            checkStations = lambda system: any(
                stn for stn in system.stations
                if stn.checkPadSize(maxPadSize)
            )
        
        while openSet:
            weight, curDist, curSysID, stnDist = heappop(openSet)
            # If we reached 'goal' we've found the shortest path.
            if curSysID == destID:
                break
            if curDist >= distTo:
                continue
            curSys = sysByID[curSysID]
            # A node might wind up multiple times on the open list,
            # so check if we've already found a shorter distance to
            # the system and if so, ignore it this time.
            if curDist > distances[curSys][1]:
                continue
            
            system_iter = iter(systemsInRange(curSys, maxJumpLy))
            if stationInterval:
                if checkStations(curSys):
                    stnDist = 0
                else:
                    stnDist += 1
                    if stnDist >= stationInterval:
                        system_iter = iter(
                            v for v in system_iter if checkStations(v[0])
                        )
            
            distFn = curSys.distanceTo
            for nSys, nDist in system_iter:
                newDist = curDist + nDist
                if getDist(nSys, defaultDist)[1] <= newDist:
                    continue
                distances[nSys] = (curSys, newDist)
                weight = distFn(nSys)
                nID = nSys.ID
                heappush(openSet, (newDist + weight, newDist, nID, stnDist))
                if nID == destID:
                    distTo = newDist
        
        if dest not in distances:
            return None
        
        path = []
        
        while True:
            (prevSys, dist) = getDist(dest)
            path.append((dest, dist))
            if dest == origin:
                break
            dest = prevSys
        
        path.reverse()
        
        return path
    
    ############################################################
    # Station data.
    
    def stations(self):
        """ Iterate through the list of stations. """
        yield from self.stationByID.values()
    
    def _loadStations(self):
        """
        Populate the Station list.
        Station constructor automatically adds itself to the System object.
        CAUTION: Will orphan previously loaded objects.
        """
        stmt = """
            SELECT  station_id, system_id, name,
                    ls_from_star, market, blackmarket, shipyard,
                    max_pad_size, outfitting, rearm, refuel, repair, planetary, type_id
              FROM  Station
        """
        self.cur.execute(stmt)
        stationByID = {}
        systemByID = self.systemByID
        self.tradingStationCount = 0
        # Fleet Carriers are station type 24.
        # Storing as a list allows easy expansion if needed.
        fc_types = [24]
        for (
            ID, systemID, name,
            lsFromStar, market, blackMarket, shipyard,
            maxPadSize, outfitting, rearm, refuel, repair, planetary, type_id
        ) in self.cur:
            isFleet = 'Y' if int(type_id) in fc_types else 'N'
            station = Station(
                ID, systemByID[systemID], name,
                lsFromStar, market, blackMarket, shipyard,
                maxPadSize, outfitting, rearm, refuel, repair, planetary, isFleet, 
                0, None,
            )
            stationByID[ID] = station
        
        tradingCount = 0
        stmt = """
            SELECT  station_id,
                    COUNT(*) AS item_count,
                    AVG(JULIANDAY('now') - JULIANDAY(modified))
              FROM  StationItem
             GROUP  BY 1
             HAVING item_count > 0
        """
        for ID, itemCount, dataAge in self.cur.execute(stmt):
            station = stationByID[ID]
            station.itemCount = itemCount
            station.dataAge = dataAge
            tradingCount += 1
        
        self.stationByID = stationByID
        self.tradingStationCount = tradingCount
        self.tdenv.DEBUG1("Loaded {:n} Stations", len(stationByID))
        self.stellarGrid = None
    
    def addLocalStation(
            self,
            system,
            name,
            lsFromStar,
            market,
            blackMarket,
            shipyard,
            maxPadSize,
            outfitting,
            rearm,
            refuel,
            repair,
            planetary,
            modified='now',
            commit=True,
            ):
        """
        Add a station to the local cache and memory copy.
        """
        
        market = market.upper()
        blackMarket = blackMarket.upper()
        shipyard = shipyard.upper()
        maxPadSize = maxPadSize.upper()
        outfitting = outfitting.upper()
        rearm = rearm.upper()
        refuel = refuel.upper()
        repair = repair.upper()
        planetary = planetary.upper()
        assert market in "?YN"
        assert blackMarket in "?YN"
        assert shipyard in "?YN"
        assert maxPadSize in "?SML"
        assert outfitting in "?YN"
        assert rearm in "?YN"
        assert refuel in "?YN"
        assert repair in "?YN"
        assert planetary in '?YN'
        
        db = self.getDB()
        cur = db.cursor()
        cur.execute("""
            INSERT INTO Station (
                name, system_id,
                ls_from_star, market, blackmarket, shipyard, max_pad_size,
                outfitting, rearm, refuel, repair, planetary,
                modified
            ) VALUES (
                ?, ?,
                ?, ?, ?, ?, ?,
                ?, ?, ?, ?, ?,
                DATETIME(?)
            )
        """, [
            name, system.ID,
            lsFromStar, market, blackMarket, shipyard, maxPadSize,
            outfitting, rearm, refuel, repair, planetary,
            modified,
        ])
        ID = cur.lastrowid
        station = Station(
            ID, system, name,
            lsFromStar=lsFromStar,
            market=market,
            blackMarket=blackMarket,
            shipyard=shipyard,
            maxPadSize=maxPadSize,
            outfitting=outfitting,
            rearm=rearm,
            refuel=refuel,
            repair=repair,
            planetary=planetary,
            fleet='?',
            itemCount=0, dataAge=0,
        )
        self.stationByID[ID] = station
        if commit:
            db.commit()
        self.tdenv.NOTE(
            "{} (#{}) added to {}: "
            "ls={}, mkt={}, bm={}, yard={}, pad={}, "
            "out={}, arm={}, ref={}, rep={}, plt={}, "
            "mod={}",
            station.name(), station.ID,
            self.dbPath if self.tdenv.detail > 1 else "local db",
            lsFromStar, market, blackMarket, shipyard, maxPadSize,
            outfitting, rearm, refuel, repair, planetary,
            modified,
        )
        return station
    
    def updateLocalStation(
            self, station,
            name=None,
            lsFromStar=None,
            market=None,
            blackMarket=None,
            shipyard=None,
            maxPadSize=None,
            outfitting=None,
            rearm=None,
            refuel=None,
            repair=None,
            planetary=None,
            modified='now',
            force=False,
            commit=True,
            ):
        """
        Alter the properties of a station in-memory and in the DB.
        """
        changes = []
        
        def _changed(label, old, new):
            changes.append(
                "{}('{}'=>'{}')".format(label, old, new)
            )
        
        if name is not None:
            if force or name.upper() != station.dbname.upper():
                _changed("name", station.dbname, name)
                station.dbname = name
        
        if lsFromStar is not None:
            assert lsFromStar >= 0
            if lsFromStar != station.lsFromStar:
                if lsFromStar > 0 or force:
                    _changed("ls", station.lsFromStar, lsFromStar)
                    station.lsFromStar = lsFromStar
        
        def _check_setting(label, name, newValue, allowed):
            if newValue is not None:
                newValue = newValue.upper()
                assert newValue in allowed
                oldValue = getattr(station, name, '?')
                if newValue != oldValue and (force or newValue != '?'):
                    _changed(label, oldValue, newValue)
                    setattr(station, name, newValue)
        
        _check_setting("pad", "maxPadSize", maxPadSize, TradeDB.padSizes)
        _check_setting("mkt", "market", market, TradeDB.marketStates)
        _check_setting("blk", "blackMarket", blackMarket, TradeDB.marketStates)
        _check_setting("shp", "shipyard", shipyard, TradeDB.marketStates)
        _check_setting("out", "outfitting", outfitting, TradeDB.marketStates)
        _check_setting("arm", "rearm", rearm, TradeDB.marketStates)
        _check_setting("ref", "refuel", refuel, TradeDB.marketStates)
        _check_setting("rep", "repair", repair, TradeDB.marketStates)
        _check_setting("plt", "planetary", planetary, TradeDB.planetStates)
        _check_setting("flc", "fleet", fleet, TradeDB.fleetStates)
        
        if not changes:
            return False
        
        db = self.getDB()
        db.execute("""
            UPDATE Station
               SET name=?,
                   ls_from_star=?,
                   market=?,
                   blackmarket=?,
                   shipyard=?,
                   max_pad_size=?,
                   outfitting=?,
                   rearm=?,
                   refuel=?,
                   repair=?,
                   planetary=?,
                   modified=DATETIME(?)
             WHERE station_id = ?
        """, [
            station.dbname,
            station.lsFromStar,
            station.market,
            station.blackMarket,
            station.shipyard,
            station.maxPadSize,
            station.outfitting,
            station.rearm,
            station.refuel,
            station.repair,
            station.planetary,
            modified,
            station.ID
        ])
        if commit:
            db.commit()
        
        self.tdenv.NOTE(
            "{} (#{}) updated in {}: {}",
            station.name(), station.ID,
            self.dbPath if self.tdenv.detail > 1 else "local db",
            ", ".join(changes)
        )
        
        return True
    
    def removeLocalStation(self, station, commit=True):
        """
        Removes a station from the local database and memory image.
        
        Becareful of any references to the station you may still have
        after this.
        """
        
        # Remove reference from my system
        system = station.system
        system.stations.remove(station)
        
        # Remove the ID lookup
        del self.stationByID[station.ID]
        
        # Delete database entry
        db = self.getDB()
        db.execute("""
            DELETE FROM Station
             WHERE system_id = ? AND station_id = ?
        """, [system.ID, station.ID]
        )
        if commit:
            db.commit()
        
        self.tdenv.NOTE(
            "{} (#{}) deleted from {}",
            station.name(), station.ID,
            self.dbPath if self.tdenv.detail > 1 else "local db",
        )
        
        station.dbname = "DELETED "+station.dbname
        del station
    
    def lookupPlace(self, name):
        """
        Lookup the station/system specified by 'name' which can be the
        name of a System or Station or it can be "System/Station" when
        the user needs to disambiguate a station. In this case, both
        system and station can be partial matches.
        
        The system tries to allow partial matches as well as matches
        which omit whitespaces. In order to do this and still support
        the massive namespace of Stars and Systems, we rank the
        matches so that exact matches win, and only inferior close
        matches are looked at if no exacts are found.
        
        Legal annotations:
            system
            station
            @system    [explicitly a system name]
            /station   [explicitly a station name]
            system/station
            @system/station
        """
        
        if isinstance(name, System) or isinstance(name, Station):
            return name
        
        slashPos = name.find('/')
        if slashPos < 0:
            slashPos = name.find('\\')
        nameOff = 1 if name.startswith('@') else 0
        if slashPos > nameOff:
            # Slash indicates it's, e.g., AULIN/ENTERPRISE
            sysName = name[nameOff:slashPos].upper()
            stnName = name[slashPos+1:]
        elif slashPos == nameOff:
            sysName, stnName = None, name[nameOff+1:]
        elif nameOff:
            # It's explicitly a station
            sysName, stnName = name[nameOff:].upper(), None
        else:
            # It could be either, use the name for both.
            stnName = name[nameOff:]
            sysName = stnName.upper()
        
        exactMatch = []
        closeMatch = []
        wordMatch = []
        anyMatch = []
        
        def lookup(name, candidates):
            """ Search candidates for the given name """
            
            normTrans = TradeDB.normalizeTrans
            trimTrans = TradeDB.trimTrans
            
            nameNorm = name.translate(normTrans)
            nameTrimmed = nameNorm.translate(trimTrans)
            
            nameLen = len(name)
            nameNormLen = len(nameNorm)
            nameTrimmedLen = len(nameTrimmed)
            
            for place in candidates:
                placeName = place.dbname
                placeNameNorm = placeName.translate(normTrans)
                placeNameNormLen = len(placeNameNorm)
                
                if nameTrimmedLen > placeNameNormLen:
                    # The needle is bigger than this haystack.
                    continue
                
                # If the lengths match, do a direct comparison.
                if len(placeName) == nameLen:
                    if placeNameNorm == nameNorm:
                        exactMatch.append(place)
                    continue
                if placeNameNormLen == nameNormLen:
                    if placeNameNorm == nameNorm:
                        closeMatch.append(place)
                    continue
                
                if nameNormLen < placeNameNormLen:
                    subPos = placeNameNorm.find(nameNorm)
                    if subPos == 0:
                        if placeNameNorm[nameNormLen] == ' ':
                            # first word
                            wordMatch.append(place)
                        else:
                            anyMatch.append(place)
                        continue
                    elif subPos > 0:
                        if placeNameNorm[subPos] == ' ' and \
                                placeNameNorm[subPos + nameNormLen] == ' ':
                            wordMatch.append(place)
                        else:
                            anyMatch.append(place)
                        continue
                
                # Lets drop whitespace and remaining punctuation...
                placeNameTrimmed = placeNameNorm.translate(trimTrans)
                placeNameTrimmedLen = len(placeNameTrimmed)
                if placeNameTrimmedLen == placeNameNormLen:
                    # No change
                    continue
                
                # A match here is not exact but still fairly interesting
                if len(placeNameTrimmed) == nameTrimmedLen:
                    if placeNameTrimmed == nameTrimmed:
                        closeMatch.append(place)
                    continue
                if placeNameTrimmed.find(nameTrimmed) >= 0:
                    anyMatch.append(place)
        
        if sysName:
            try:
                sys = self.systemByName[sysName]
                exactMatch = [sys]
            except KeyError:
                lookup(sysName, self.systemByID.values())
        if stnName:
            # Are we considering the name as a station?
            # (we don't if they type, e,g '@aulin')
            # compare against nameOff to allow '@/station'
            if slashPos > nameOff + 1:
                # "sys/station"; the user should have specified a system
                # name and we should be able to narrow down which
                # stations we compare against. Check first if there are
                # any matches.
                stationCandidates = []
                for sys in itertools.chain(
                        exactMatch, closeMatch, wordMatch, anyMatch
                        ):
                    stationCandidates += sys.stations
                # Clear out the candidate lists
                exactMatch = []
                closeMatch = []
                wordMatch = []
                anyMatch = []
            else:
                # Consider against all station names
                stationCandidates = self.stationByID.values()
            lookup(stnName, stationCandidates)
        
        # consult the match sets in ranking order for a single
        # match, which denotes a win at that tier. For example,
        # if there is one exact match, we don't care how many
        # close matches there were.
        for matchSet in exactMatch, closeMatch, wordMatch, anyMatch:
            if len(matchSet) == 1:
                return matchSet[0]
        
        # Nothing matched
        if not any([exactMatch, closeMatch, wordMatch, anyMatch]):
            # Note: this was a TradeException and may need to be again,
            # but then we need to catch that error in commandenv
            # when we process avoids
            raise LookupError("Unrecognized place: {}".format(name))
        
        # More than one match
        raise AmbiguityError(
            'System/Station', name,
            exactMatch + closeMatch + wordMatch + anyMatch,
            key=lambda place: place.name()
        )
    
    def lookupStation(self, name, system=None):
        """
        Look up a Station object by it's name or system.
        """
        if isinstance(name, Station):
            return name
        if isinstance(name, System):
            # When given a system with only one station, return the station.
            if len(name.stations) != 1:
                raise SystemNotStationError(
                    "System '%s' has %d stations, "
                    "please specify a station instead." % (
                        name.str(), len(name.stations)
                    )
                )
            return name.stations[0]
        
        if system:
            system = self.lookupSystem(system)
            return TradeDB.listSearch(
                "Station", name, system.stations,
                key=lambda system: system.dbname)
        
        stationID, station, systemID, system = None, None, None, None
        try:
            system = TradeDB.listSearch(
                "System", name, self.systemByID.values(),
                key=lambda system: system.dbname
            )
        except LookupError:
            pass
        try:
            station = TradeDB.listSearch(
                "Station", name, self.stationByID.values(),
                key=lambda station: station.dbname
            )
        except LookupError:
            pass
        # If neither matched, we have a lookup error.
        if not (station or system):
            raise LookupError(
                "'%s' did not match any station or system." % (name)
            )
        
        # If we matched both a station and a system, make sure they resovle to
        # the same station otherwise we have an ambiguity. Some stations have
        # the same name as their star system (Aulin/Aulin Enterprise)
        if system and station and system != station.system:
            raise AmbiguityError(
                'Station', name, [system.name(), station.name()]
            )
        
        if station:
            return station
        
        # If we only matched a system name, ensure that it's a single station
        # system otherwise they need to specify a station name.
        if len(system.stations) != 1:
            raise SystemNotStationError(
                "System '%s' has %d stations, "
                "please specify a station instead." % (
                    system.name(), len(system.stations)
                )
            )
        return system.stations[0]
    
    def getDestinations(
            self,
            origin,
            maxJumps=None,
            maxLyPer=None,
            avoidPlaces=None,
            maxPadSize=None,
            maxLsFromStar=0,
            noPlanet=False,
            planetary=None,
            fleet=None,
            ):
        """
        Gets a list of the Station destinations that can be reached
        from this Station within the specified constraints.
        Limits to stations we are trading with if trading is True.
        """
        
        if maxJumps is None:
            maxJumps = sys.maxsize
        maxLyPer = maxLyPer or self.maxSystemLinkLy
        if avoidPlaces is None:
            avoidPlaces = ()
        
        # The open list is the list of nodes we should consider next for
        # potential destinations.
        # The path list is a list of the destinations we've found and the
        # shortest path to them. It doubles as the "closed list".
        # The closed list is the list of nodes we've already been to (so
        # that we don't create loops A->B->C->A->B->C->...)
        
        origSys = origin.system if isinstance(origin, Station) else origin
        openList = [DestinationNode(origSys, [origSys], 0)]
        # I don't want to have to consult both the pathList
        # AND the avoid list every time I'm considering a
        # station, so copy the avoid list into the pathList
        # with a negative distance so I can ignore them again
        # when I scrape the pathList.
        # Don't copy stations because those only affect our
        # termination points, and not the systems we can
        # pass through en-route.
        pathList = {
            system.ID: DestinationNode(system, None, -1.0)
            for system in avoidPlaces
            if isinstance(system, System)
        }
        if origSys.ID not in pathList:
            pathList[origSys.ID] = openList[0]
        
        # As long as the open list is not empty, keep iterating.
        jumps = 0
        while openList and jumps < maxJumps:
            # Expand the search domain by one jump; grab the list of
            # nodes that are this many hops out and then clear the list.
            ring, openList = openList, []
            # All of the destinations we are about to consider will
            # either be on the closed list or they will be +1 jump away.
            jumps += 1
            
            ring.sort(key=lambda dn: dn.distLy)
            
            for node in ring:
                for (destSys, destDist) in self.genSystemsInRange(
                        node.system, maxLyPer, False
                        ):
                    dist = node.distLy + destDist
                    # If we already have a shorter path, do nothing
                    try:
                        prevDist = pathList[destSys.ID].distLy
                    except KeyError:
                        pass
                    else:
                        if dist >= prevDist:
                            continue
                    # Add to the path list
                    destNode = DestinationNode(
                        destSys, node.via + [destSys], dist
                    )
                    pathList[destSys.ID] = destNode
                    # Add to the open list but also include node to the via
                    # list so that it serves as the via list for all next-hops.
                    openList.append(destNode)
        
        # We have a system-to-system path list, now we
        # need stations to terminate at.
        def path_iter_fn():
            for node in pathList.values():
                if node.distLy >= 0.0:
                    for station in node.system.stations:
                        yield node, station
        
        path_iter = iter(path_iter_fn())
        if noPlanet:
            path_iter = iter(
                (node, station) for (node, station) in path_iter
                if station.planetary == 'N'
            )
        if avoidPlaces:
            path_iter = iter(
                (node, station) for (node, station) in path_iter
                if station not in avoidPlaces
            )
        if maxPadSize:
            path_iter = iter(
                (node, station) for (node, station) in path_iter
                if station.checkPadSize(maxPadSize)
            )
        if planetary:
            path_iter = iter(
                (node, station) for (node, station) in path_iter
                if station.checkPlanetary(planetary)
            )
        if fleet:
            path_iter = iter(
                (node, station) for (node, station) in path_iter
                if station.checkFleet(fleet)
            )
        if maxLsFromStar:
            path_iter = iter(
                (node, stn) for (node, stn) in path_iter
                if stn.lsFromStar > 0 and stn.lsFromStar <= maxLsFromStar
            )
        for node, stn in path_iter:
            yield Destination(node.system, stn, node.via, node.distLy)
    
    ############################################################
    # Ship data.
    
    def ships(self):
        """ Iterate through the list of ships. """
        yield from self.shipByID.values()
    
    def _loadShips(self):
        """
        Populate the Ship list.
        CAUTION: Will orphan previously loaded objects.
        """
        stmt = """
            SELECT ship_id, name, cost, fdev_id
              FROM Ship
        """
        self.cur.execute(stmt)
        self.shipByID = {
            row[0]: Ship(*row, stations=[])
            for row in self.cur
        }
        
        self.tdenv.DEBUG1("Loaded {} Ships", len(self.shipByID))
    
    def lookupShip(self, name):
        """
        Look up a ship by name
        """
        return TradeDB.listSearch(
            "Ship", name, self.shipByID.values(),
            key=lambda ship: ship.dbname
        )
    
    ############################################################
    # Item data.
    
    def categories(self):
        """
        Iterate through the list of categories.
        key = category name, value = list of items.
        """
        yield from self.categoryByID.items()
    
    def _loadCategories(self):
        """
        Populate the list of item categories.
        CAUTION: Will orphan previously loaded objects.
        """
        stmt = """
            SELECT category_id, name
              FROM Category
        """
        self.categoryByID = {
            ID: Category(ID, name, [])
            for (ID, name) in self.cur.execute(stmt)
        }
        
        self.tdenv.DEBUG1("Loaded {} Categories", len(self.categoryByID))
    
    def lookupCategory(self, name):
        """
        Look up a category by name
        """
        return TradeDB.listSearch(
            "Category", name,
            self.categoryByID.values(),
            key=lambda cat: cat.dbname
        )
    
    def items(self):
        """ Iterate through the list of items. """
        yield from self.itemByID.values()
    
    def _loadItems(self):
        """
        Populate the Item list.
        CAUTION: Will orphan previously loaded objects.
        """
        stmt = """
            SELECT item_id, name, category_id, avg_price, fdev_id
              FROM Item
        """
        itemByID, itemByName, itemByFDevID = {}, {}, {}
        for ID, name, categoryID, avgPrice, fdevID in self.cur.execute(stmt):
            category = self.categoryByID[categoryID]
            item = Item(
                ID, name, category,
                '{}/{}'.format(category.dbname, name),
                avgPrice, fdevID
            )
            itemByID[ID] = item
            itemByName[name] = item
            if fdevID:
                itemByFDevID[fdevID] = item
            
            category.items.append(item)
        
        self.itemByID = itemByID
        self.itemByName = itemByName
        self.itemByFDevID = itemByFDevID
        
        self.tdenv.DEBUG1(
            "Loaded {:n} Items",
            len(self.itemByID)
        )
    
    def lookupItem(self, name):
        """
            Look up an Item by name using "CATEGORY/Item"
        """
        return TradeDB.listSearch(
            "Item", name, self.itemByName.items(),
            key=lambda kvTup: kvTup[0],
            val=lambda kvTup: kvTup[1]
        )
    
    def getAverageSelling(self):
        """
        Query the database for average selling prices of all items.
        """
        if not self.avgSelling:
            self.avgSelling = {itemID: 0 for itemID in self.itemByID.keys()}
            self.avgSelling.update({
                ID: int(cr)
                for ID, cr in self.getDB().execute("""
                    SELECT  i.item_id, IFNULL(AVG(supply_price), 0)
                      FROM  Item AS i
                            LEFT OUTER JOIN StationItem AS si ON (
                                i.item_id = si.item_id AND si.supply_price > 0
                            )
                     WHERE  supply_price > 0
                     GROUP  BY 1
                """)
            })
        return self.avgSelling
    
    def getAverageBuying(self):
        """
        Query the database for average buying prices of all items.
        """
        if not self.avgBuying:
            self.avgBuying = {itemID: 0 for itemID in self.itemByID.keys()}
            self.avgBuying.update({
                ID: int(cr)
                for ID, cr in self.getDB().execute("""
                    SELECT  i.item_id, IFNULL(AVG(demand_price), 0)
                      FROM  Item AS i
                            LEFT OUTER JOIN StationItem AS si ON (
                                i.item_id = si.item_id AND si.demand_price > 0
                            )
                     WHERE  demand_price > 0
                     GROUP  BY 1
                """)
            })
        return self.avgBuying
    
    ############################################################
    # Rare Items
    
    def _loadRareItems(self):
        """
        Populate the RareItem list.
        """
        stmt = """
            SELECT  rare_id, station_id, category_id, name,
                    cost, max_allocation, illegal, suppressed
              FROM  RareItem
        """
        self.cur.execute(stmt)
        
        rareItemByID, rareItemByName = {}, {}
        stationByID = self.stationByID
        for (
            ID, stnID, catID, name,
            cost, maxAlloc, illegal, suppressed
        ) in self.cur:
            station = stationByID[stnID]
            category = self.categoryByID[catID]
            rare = RareItem(
                ID, station, name, cost, maxAlloc, illegal, suppressed,
                category, '{}/{}'.format(category.dbname, name)
            )
            rareItemByID[ID] = rareItemByName[name] = rare
        self.rareItemByID = rareItemByID
        self.rareItemByName = rareItemByName
        
        self.tdenv.DEBUG1(
            "Loaded {:n} RareItems",
            len(rareItemByID)
        )
    
    ############################################################
    # Price data.
    
    def close(self):
        self.cur = None
        if self.conn:
            self.conn.close()
        self.conn = None
    
    def load(self, maxSystemLinkLy=None):
        """
            Populate/re-populate this instance of TradeDB with data.
            WARNING: This will orphan existing records you have
            taken references to:
                tdb.load()
                x = tdb.lookupPlace("Aulin")
                tdb.load() # x now points to an orphan Aulin
        """
        
        self.tdenv.DEBUG1("Loading data")
        
        self.conn = conn = self.getDB()
        self.cur = conn.cursor()
        
        self._loadAdded()
        self._loadSystems()
        self._loadStations()
        self._loadShips()
        self._loadCategories()
        self._loadItems()
        self._loadRareItems()
        
        # Calculate the maximum distance anyone can jump so we can constrain
        # the maximum "link" between any two stars.
        msll = maxSystemLinkLy or self.tdenv.maxSystemLinkLy or 30
        self.maxSystemLinkLy = msll
    
    ############################################################
    # General purpose static methods.
    
    @staticmethod
    def listSearch(
            listType, lookup, values,
            key=lambda item: item,
            val=lambda item: item
            ):
        """
        Searches [values] for 'lookup' for least-ambiguous matches,
        return the matching value as stored in [values].
        
        GIVEN [values] contains "bread", "water", "biscuits and "It",
        searching "ea" will return "bread", "WaT" will return "water"
        and "i" will return "biscuits".
        
        Searching for "a" would raise an AmbiguityError because "a" matches
        "bread" and "water", but searching for "it" will return "It"
        because it provides an exact match of a key.
        """
        
        class ListSearchMatch(namedtuple('Match', ['key', 'value'])):
            pass
        
        normTrans = TradeDB.normalizeTrans
        trimTrans = TradeDB.trimTrans
        needle = lookup.translate(normTrans).translate(trimTrans)
        partialMatch, wordMatch = [], []
        # make a regex to match whole words
        wordRe = re.compile(r'\b{}\b'.format(lookup), re.IGNORECASE)
        # describe a match
        for entry in values:
            entryKey = key(entry)
            normVal = entryKey.translate(normTrans).translate(trimTrans)
            if normVal.find(needle) > -1:
                # If this is an exact match, ignore ambiguities.
                if len(normVal) == len(needle):
                    return val(entry)
                match = ListSearchMatch(entryKey, val(entry))
                if wordRe.match(entryKey):
                    wordMatch.append(match)
                else:
                    partialMatch.append(match)
        # Whole word matches trump partial matches
        if wordMatch:
            if len(wordMatch) > 1:
                raise AmbiguityError(
                    listType, lookup, wordMatch,
                    key=lambda item: item.key,
                )
            return wordMatch[0].value
        # Fuzzy matches
        if partialMatch:
            if len(partialMatch) > 1:
                raise AmbiguityError(
                    listType, lookup, partialMatch,
                    key=lambda item: item.key,
                )
            return partialMatch[0].value
        # No matches
        raise LookupError(
            "Error: '%s' doesn't match any %s" % (lookup, listType)
        )
    
    @staticmethod
    def normalizedStr(text):
        """
            Returns a case folded, sanitized version of 'str' suitable for
            performing simple and partial matches against. Removes various
            punctuation characters that don't contribute to name uniqueness.
            NOTE: No-longer removes whitespaces or apostrophes.
        """
        return text.translate(
            TradeDB.normalizeTrans
        ).translate(
            TradeDB.trimTrans
        )

######################################################################
# Assorted helpers

def describeAge(ageInSeconds):
    """
    Turns an age (in seconds) into a text representation.
    """
    hours = int(ageInSeconds / 3600)
    if hours < 1:
        return "<1 hr"
    if hours == 1:
        return "1 hr"
    if hours < 48:
        return str(hours) + " hrs"
    days = int(hours / 24)
    if days < 90:
        return str(days) + " days"
    
    months = int(days / 31)
    return str(months) + " mths"



#!/usr/bin/env python3.6

"""
This tool looks for changes in the EDSC service since the most
recent "modified" date in the System table or the date supplied
on the command line.

It then tries to do some validation but also requires user
confirmation.

For each star that appears to be new, it copies the name into
the clipboard so you can paste it into the "SEARCH" box in the
game to verify that the name is correct.

Additionally it shows you the distance from "current system"
to the star as a way to verify the co-ordinates.

This helps to catch cases where people have typo'd system names,
but given the right coordinates; it also helps catch cases where
people have used the star name from in-system which sometimes
differs from the star name in the galaxy map.

For each star you can type "y" to accept the star, "n" to skip it
or "q" to stop recording.
"""

import argparse
import math
import misc.clipboard
import misc.edsc
import os
import random
import re
import sys
import tradedb
import tradeenv

DEFAULT_DATE = "2015-04-01 00:00:00"

# Systems we know are bad.
ignore = [
    "ADAM NAPAT",
    "ALRAI SECTOR IR-W C1-29",
    "ALRAI SECTOR OI-T B3-6 A",
    "AN SEXSTANS",
    "ARIETIS SECTOR WO-R 84-4",
    "COL 285 SECTOR GI-P B20-4 A",
    "COL 285 SECTOR GU-T 818-2",
    "COL 285 SECTOR GU-T 818-5",
    "COL 285 SECTOR HU-T 818-1",
    "COL 285 SECTOR ZJ-Y B140-1",
    "CORE SYS SECTOR DL-Y D117",
    "CORE SYS SECTOR HH-V B2-7",
    "CRU7CIS SECTOR EQ-Y B2",
    "CRUCIS SECTO GB-X B1-0",
    "CRUCIS SECTOR FM-V B2-0",
    "CRUCIS SECTOR YE-A D142",
    "CRUSCIC SECTOR EL-Y C19",
    "CRUSIS SECTOR MD-S B4-3",
    "CRUSIS SECTOR ND-S B4-1",
    "DQ TUCANE",
    "ED TUCANAE",
    "FLECHS",
    "GD 139",
    "HYADES SECTOR WF-M B8-3",
    "22 LYNCIS",
    "ALANI",
    "BODB DJEDI",
    "DJALI",
    "HIP 101110",
    "NJUNG",
    "PANTAA CEZISA",
    "TAVYTERE",
    "DITIBI",
    "COL 285 SECTOR EC-R B18-5",
    "DITIBI (FIXED)",
    "HYADES",
    "HAREMID",
    "HYADES SECTOR VO-Q V5-1",
    "XI HUGAN",
    "LP  634-18",
    "SHU WEI SECTOR MN-S B4-4",
    "SHU WEI SECTOR MN-S B4-9",
    "THETA CARINE",
    "CORE SYS HH-M A7-3",
    "OI-T B3-9",
    "IDZ DL-X B1-1",
    "PHE ZHUA",
    "LP 937-98",
    "WREGOE QC-X B45-2",
    "ALRAI SECTOR JW-B A6-1",
    "BEI DOU SECTOR JV-C B2-2",
    "BLANCO 1 SECTOR EG-Y D8",
    "HERCULIS SECTOR ZE-A B86",
    "LFT 16653",
    "LP 649-73",
    "OMAGATE",
    "PISCIUM SECTOR YU-Y C171",
    "PLAA EURK AX-A B0",
    "PRAEA EUQ GR-O B13-0",
    "PRAEA EUQ GV-W C",
    "PRAEA EUQ SX-U S2-18",
    "SYNUEFAI FL-V D3-7",
    "SYNUEFAI NM-",
    "COL 285 SECTOR BG-Y B16-1",
    "SYNUEFE NA-O D13-16",
    "HIP 623536",
    "SYNUEFUE D2-26",
    "WREGOE AZ-D B45-4",
    "MINIMAR",
    "MV URSAE MAJORIS",
    "PLAA EURK BU-Q B5-0",
    "COL 285 SECTOR LP-S B-19-1",
    "COL 285 SECTOR LP-S B-19-2",
    "CORE SYS SECTOR CB-0 A6-5",
    "ARIETIS SECTOR JD-F A12-3",
    "ICZ-KI-S B4-3",
    "HIP 8153",
    "HYADES SECTOR MS-X B4-4",
    "UCAC3 70-2386",
    "HYADES KX-T C3-24",
    "ICZ ZF-O B2-6",
    "ZELAND",
]

class UsageError(Exception):
    """ Raised when command line usage is invalid. """
    pass

def parse_arguments():
    parser = argparse.ArgumentParser(
            description='Review and validate incoming EDSC star data. '
                        'Confirmed distances are submitted back to EDSC to '
                        'increase confidence ratings.',
            epilog='Confirmed systems are written to tmp/new.systems.csv.',
    )
    parser.add_argument(
            'refSystem',
            help='*Exact* name of the system you are *currently* in, '
                 'used as a reference system for distance validations.',
            type=str,
            metavar='"REFERENCE SYSTEM"',
            default=None,
            nargs='?',
    )
    parser.add_argument(
            '--cmdr',
            required=False,
            help='Specify your commander name.',
            type=str,
            default=os.environ.get('CMDR', None),
    )
    parser.add_argument(
            '--confidence',
            required=False,
            help='Specify minimum confidence level.',
            type=int,
            default=2,
    )
    grp = parser.add_mutually_exclusive_group()
    if grp: # for indentation
        grp.add_argument(
                '--random',
                action='store_true',
                required=False,
                help='Show systems in random order, maximum of --max-systems.',
        )
        grp.add_argument(
                '--distance',
                action='store_true',
                required=False,
                help='Select upto 10 systems by proximity.',
        )
    parser.add_argument(
            '--max-systems',
            dest='maxSystems',
            help='Maximum systems to query with --random/--distance.',
            required=False,
            type=int,
            default=10
        )
    parser.add_argument(
            '--add-to-local-db', '-A',
            action='store_true',
            required=False,
            help='Add accepted systems to the local database.',
            dest='add',
    )
    parser.add_argument(
            '--test',
            required=False,
            help='Use the EDSC test database.',
            action='store_true',
            default=False,
    )
    parser.add_argument(
            '--date',
            required=False,
            help='Use specified date (YYYY-MM-DD HH:MM:SS format) for '
                 'start of update search. '
                 'Default is to use the last System modified date.',
            type=str,
            default=None,
    )
    parser.add_argument(
            '--no-splash', '-NS',
            required=False,
            action='store_false',
            help="Don't display the 'splash' text on startup.",
            dest='splash',
            default=True,
    )
    parser.add_argument(
            '--summary',
            required=False,
            help='Check for and report on new systems but do no work.',
            action='store_true',
    )
    parser.add_argument(
            '--detail', '-v',
            help='Increase level  of detail in output.',
            default=0,
            required=False,
            action='count',
    )
    parser.add_argument(
            '--debug', '-w',
            help='Enable/raise level of diagnostic output.',
            default=0,
            required=False,
            action='count',
    )
    parser.add_argument(
            '--ref',
            help='Reference system (for --distance).',
            default=None,
            dest='refSys',
            type=str,
    )
    
    argv = parser.parse_args(sys.argv[1:])
    dateRe = re.compile(r'^20\d\d-(0\d|1[012])-([012]\d|3[01]) ([01]\d|2[0123]):[0-5]\d:[0-5]\d$')
    if not argv.summary:
        if not argv.cmdr:
            raise UsageError("No --cmdr specified / no CMDR environment variable")
        if not argv.refSystem:
            raise UsageError(
                "Must specify a reference system name (when not using "
                "--summary). Be sure to put the name in double quotes, "
                "e.g. \"SOL\" or \"I BOOTIS\".\n"
            )
    if not argv.date:
        argv.date = DEFAULT_DATE
    if not dateRe.match(argv.date):
        raise UsageError(
                "Invalid date: '{}', expecting YYYY-MM-DD HH:MM:SS format."
                .format(argv.date)
        )
    if not argv.distance and argv.refSys:
        raise UsageError("--ref requires --distance")
    
    return argv

def is_change(tdb, sysinfo):
    """ Check if a system's EDSC data is different than TDs """
    name = sysinfo['name'] = sysinfo['name'].upper()
    if name.startswith("PLAA EURIK"):
        return False
    if name in ignore:
        return False
    try:
        x, y, z = sysinfo['coord']
        place = tdb.systemByName[name]
        if place.posX == x and place.posY == y and place.posZ == z:
            return False
    except KeyError:
        place = None
    sysinfo['place'] = place
    return True

def has_position_changed(place, name, x, y, z):
    if not place:
        return False
    
    print("! @{} [{},{},{}] changed to @{} [{},{},{}]".format(
        name, x, y, z,
        place.dbname, place.posX, place.posY, place.posZ
    ))
    
    return True

def check_database(tdb, name, x, y, z):
    # is it in the database?
    cur = tdb.query("""
        SELECT name, pos_x, pos_y, pos_z
          FROM System
         WHERE pos_x BETWEEN ? and ?
           AND pos_y BETWEEN ? and ?
           AND pos_z BETWEEN ? and ?
    """, [ 
        x - 0.5, x + 0.5,
        y - 0.5, y + 0.5,
        z - 0.5, z + 0.5,
    ])
    for mname, mx, my, mz in cur:
        print(
                "! @{} [{},{},{}] matches coords for "
                "@{} [{},{},{}]".format(
                    name, x, y, z,
                    mname, mx, my, mz
        ), file=sys.stderr)

def get_distance(tdb, startSys, x, y, z):
    distance = tdb.calculateDistance(
        startSys.posX, startSys.posY, startSys.posZ,
        x, y, z
    )
    return float("{:.2f}".format(distance))

def submit_distance(argv, name, distance, refSys=None, refDist=None):
    p0 = name.upper()
    ref = argv.startSys.name().upper()
    print("Sending: {}->{}: {}ly by {}".format(
        p0, ref, distance, argv.cmdr
    ))
    sub = misc.edsc.StarSubmission(
        star=p0,
        distances={ref: distance},
        commander=argv.cmdr,
        test=argv.test,
    )
    if refSys and refDist:
        if isinstance(refSys, tradedb.System):
            refName = refSys.name()
        else:
            refName = refSys
        sub.add_distance(refName, refDist)
    r = sub.submit()
    result = misc.edsc.StarSubmissionResult(
        star=name.upper(),
        response=r,
    )
    print(str(result))

def get_extras():
    extras = set()
    try:
        with open("data/extra-stars.txt", "rU", encoding="utf-8") as fh:
            for line in fh:
                name = line.partition('#')[0].strip().upper()
                if name:
                    extras.add(name)
    except FileNotFoundError:
        pass
    return extras

def add_to_extras(argv, name):
    with open("data/extra-stars.txt", "a", encoding="utf-8") as fh:
        print(name.upper(), file=fh)
        print("Added {} to data/extra-stars.txt".format(name))

def main():
    argv = parse_arguments()
    tdenv = tradeenv.TradeEnv(properties=argv)
    tdb = tradedb.TradeDB(tdenv)
    
    if not argv.summary:
        try:
            argv.startSys = tdb.lookupSystem(argv.refSystem)
        except (LookupError, tradedb.AmbiguityError):
            raise UsageError(
                "Unrecognized system '{}'. Reference System must be an "
                "*exact* name for a system that TD already knows.\n"
                "Did you forget to put double-quotes around the reference "
                "system name?"
                .format(argv.refSystem)
            )
    
    print("start date: {}".format(argv.date))
    
    edsq = misc.edsc.StarQuery(
        test=argv.test,
        confidence=argv.confidence,
        date=argv.date,
        )
    data = edsq.fetch()
    
    if edsq.status['statusnum'] != 0:
        raise Exception("Query failed: {} ({})".format(
                    edsq.status['msg'],
                    edsq.status['statusnum'],
                ))
    
    date = data['date']
    systems = data['systems']
    
    print("{} results".format(len(systems)))
    # Filter out systems we already know that match the EDSC data.
    systems = [
        sysinfo for sysinfo in systems if is_change(tdb, sysinfo)
    ]
    print("{} deltas".format(len(systems)))
    
    if argv.summary or len(systems) <= 0:
        return
    
    systems = [
        sysinfo for sysinfo in systems if 'coord' in sysinfo
    ]
    
    if argv.random:
        num = min(len(systems), argv.maxSystems)
        systems = random.sample(systems, num)
    
    if argv.refSys:
        refSys = tdb.lookupPlace(argv.refSys)
    else:
        refSys = None
    startSys = argv.startSys
    for sysinfo in systems:
        x, y, z = sysinfo['coord']
        sysinfo['distance'] = get_distance(tdb, startSys, x, y, z)
        if refSys:
            sysinfo['refdist'] = get_distance(tdb, refSys, x, y, z)
        else:
            sysinfo['refdist'] = None
    
    if argv.distance:
        if refSys:
            systems.sort(key=lambda sysinfo: sysinfo['refdist'])
        else:
            systems.sort(key=lambda sysinfo: sysinfo['distance'])
        systems = systems[:argv.maxSystems]
    
    if argv.splash:
        print(
            "\n"
            "===============================================================\n"
            "\n"
            " The tool will now take you through the stars returned by EDSC\n"
            " that are new or different from your local System.csv.\n"
            "\n"
            " You will be prompted with the name and predicted distance from\n"
            " your current system so you can check for mistakes.\n"
            "\n"
            " The name will be copied into your clipboard so you can alt-tab\n"
            " into the game and paste the name into the Galaxy Map's SEARCH\n"
            " box (under NAVIGATION). Let the map zoom to the system.\n"
            "\n"
            " Check the name and distance, then use the appropriate action.\n"
            "\n"
            " (Use the -NS option to skip this text in future)\n"
            "\n"
            "===============================================================\n"
            "\n"
        )
        
        input("Hit enter to continue: ")
    
    print("""At the prompt enter:
  q
      to indicate you've suffered enough,
  y
      to accept the name/value and confirm with EDSC,
  n       (or anything else)
      to skip the name/value (no confirmation),
  =name   (e.g. =SOL)
      to accept the distance but correct spelling,
  ~dist   (e.g. ~104.49)
      to submit a distance correction for the system,
""")
    print()
    
    extras = get_extras()
    
    clip = misc.clipboard.SystemNameClip()
    total = len(systems)
    current = 0
    with open("tmp/new.systems.csv", "w", encoding="utf-8") as output:
        for sysinfo in systems:
            current += 1
            name = sysinfo['name']
            created = sysinfo['createdate']
            x, y, z = sysinfo['coord']
            
            print(
                "\n"
                "-----------------------------------------------\n"
                "{syidlab:.<12}: {syid}\n"
                "{conflab:.<12}: {conf}\n"
                "{crealab:.<12}: {crcm} {crts}\n"
                "{updtlab:.<12}: {upcm} {upts}\n"
                .format(
                    syidlab="ID",
                    conflab="Confidence",
                    crealab="Created",
                    updtlab="Updated",
                    syid=sysinfo['id'],
                    conf=sysinfo['cr'],
                    crcm=sysinfo['commandercreate'],
                    crts=created,
                    upcm=sysinfo.get('commanderupdate', '[never]'),
                    upts=sysinfo.get('updatedate', ''),
                )
            )
            if refSys:
                print("{reflab:.<12}: {refdist}ly\n".format(
                    reflab="Ref Dist",
                    refdist=sysinfo['refdist'],
                ))
            
            check_database(tdb, name, x, y, z)
            
            change = has_position_changed(sysinfo['place'], name, x, y, z)
            if change:
                oldDist = startSys.distanceTo(sysinfo['place'])
                print("Old Distance: {:.2f}ly".format(oldDist))
            
            distance = sysinfo['distance']
            clip.copy_text(name)
            prompt = "{}/{}: '{}': {:.2f}ly? ".format(
                current, total,
                name,
                distance,
            )
            ok = input(prompt)
            if ok.lower() == 'q':
                break
            if ok.startswith('~'):
                correction = float(ok[1:])
                submit_distance(argv, name, correction)
                if not name.upper() in extras:
                    add_to_extras(argv, name)
                continue
            if ok.startswith('='):
                name = ok[1:].strip().upper()
                if not name in extras:
                    add_to_extras(argv, name)
                ok = 'y'
            if ok.lower() != 'y':
                continue
            
            if argv.add:
                tdb.addLocalSystem(
                    name,
                    x, y, z,
                    added='Release 1.00-EDStar',
                    modified=created,
                    commit=True
                )
            
            print("'{}',{},{},{},'Release 1.00-EDStar','{}'".format(
                name, x, y, z, created,
            ), file=output)
            
            submit_distance(argv, name, distance)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("^C")
    except UsageError as e:
        print("ERROR: {}\nSee {} --help for usage help.".format(
            str(e), sys.argv[0]
        ))



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018, 2019
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Command Line App :: Main Module
#
# TradeDangerous is a powerful set of tools for traders in Frontier
# Development's game "Elite: Dangerous". It's main function is
# calculating the most profitable trades between either individual
# stations or working out "profit runs".
#
# I wrote TD because I realized that the best trade run - in terms
# of the "average profit per stop" was rarely as simple as going
# Chango -> Dahan -> Chango.
#
# E:D's economy is complex; sometimes you can make the most profit
# by trading one item A->B and flying a second item B->A.
# But more often you need to fly multiple stations, especially since
# as you are making money different trade options are coming into
# your affordable range.
#
# END USERS: If you are a user looking to find out how to use TD,
# please consult the file "README.md".
#
# DEVELOPERS: If you are a programmer who wants TD to do something
# cool, please see the TradeDB and TradeCalc modules. TD is designed
# to empower other programmers to do cool stuff.

from __future__ import absolute_import
from __future__ import with_statement
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
import os
import sys
import traceback
import threading

from pathlib import Path
from appJar import gui
import appJar
from . import commands
from . import plugins
from .commands import exceptions
from .version import __version__

from . import tradedb
from .plugins import PluginException

WIDGET_NAMES = appJar.appjar.WIDGET_NAMES
WidgetManager = appJar.appjar.WidgetManager
# Plugins available to the 'import' command are stored here.
# The list is populated by scanning the plugin folder directly,
# so it updates automagically at start as plugins are added or removed.
#
# Any other command with available plugins must have a similar list.
importPlugs = [ plug.name[0:plug.name.find('_plug.py')]
             for plug in os.scandir(sys.modules['tradedangerous.plugins'].__path__[0])
             if plug.name.endswith("_plug.py")
             ]


def get(self, widgetType, title):
    return eval('self.get' + str(widgetType) + '("' + str(title) + '")')


gui.get = get


# @Override
def _populateSpinBox(self, spin, vals, reverse = True):
    # make sure it's a list
    #  reverse it, so the spin box functions properly
    # if reverse:
    #    vals = list(vals)
    #    vals.reverse()
    vals = tuple(vals)
    spin.config(values = vals)


gui._populateSpinBox = _populateSpinBox


# @Override
def setSpinBoxPos(self, title, pos, callFunction = True):
    spin = self.widgetManager.get(WIDGET_NAMES.SpinBox, title)
    vals = spin.cget("values")  # .split()
    vals = self._getSpinBoxValsAsList(vals)
    pos = int(pos)
    if pos < 0 or pos >= len(vals):
        raise Exception("Invalid position: " + str(pos) + ". No position in SpinBox: " + 
                    title + "=" + str(vals))
    #    pos = len(vals) - 1 - pos
    val = vals[pos]
    self._setSpinBoxVal(spin, val, callFunction)


gui.setSpinBoxPos = setSpinBoxPos


# @Override
def _configOptionBoxList(self, title, options, kind):
    """ Tidies up the list provided when an OptionBox is created/changed
    
    :param title: the title for the OptionBox - only used by TickOptionBox to calculate max size
    :param options: the list to tidy
    :param kind: The kind of option box (normal or ticks)
    :returns: a tuple containing the maxSize (width) and tidied list of items
    """
    
    # deal with a dict_keys object - messy!!!!
    if not isinstance(options, list):
        options = list(options)
    
    # make sure all options are strings
    options = [str(i) for i in options]
    
    # check for empty strings, replace first with message, remove rest
    # NO. Let's not do this annoying thing.
#    found = False
#    newOptions = []
#    for pos, item in enumerate(options):
#        if str(item).strip() == "":
#            if not found:
#                newOptions.append("- options -")
#                found = True
#        else:
#            newOptions.append(item)
#    
#    options = newOptions
    
    # get the longest string length
    try:
        maxSize = len(str(max(options, key = len)))
    except:
        try:
            maxSize = len(str(max(options)))
        except:
            maxSize = 0
    
    # increase if ticks
    if kind == "ticks":
        if len(title) > maxSize:
            maxSize = len(title)
    
    # new bug?!? - doesn't fit anymore!
    if self.platform == self.MAC:
        maxSize += 3
    return maxSize, options


gui._configOptionBoxList = _configOptionBoxList


def getOptionBox(self, title):
    """ Gets the selected item from the named OptionBox
    
    :param title: the OptionBox to check
    :returns: the selected item in an OptionBox or a dictionary of all items and their status for a TickOptionBox
    :raises ItemLookupError: if the title can't be found
    """
    box = self.widgetManager.get(WIDGET_NAMES.OptionBox, title)
    
    if box.kind == "ticks":
        val = self.widgetManager.get(WIDGET_NAMES.TickOptionBox, title, group = WidgetManager.VARS)
        retVal = {}
        for k, v in val.items():
            retVal[k] = bool(v.get())
        return retVal
    else:
        val = self.widgetManager.get(WIDGET_NAMES.OptionBox, title, group = WidgetManager.VARS).get()
#        val = val.get().strip()
        # set to None if it's a divider
#        if val.startswith("-") or len(val) == 0:
#            val = None
        return val


gui.getOptionBox = getOptionBox


def main(argv = None):
    
    class IORedirector(object):
        
        def __init__(self, TEXT_INFO):
            self.TEXT_INFO = TEXT_INFO
    
    class StdoutRedirector(IORedirector):
        
        def write(self, string):
            self.TEXT_INFO.config(text = self.TEXT_INFO.cget('text').rsplit('\r', 1)[0] + string)
        
        def flush(self):
            
            sys.__stdout__.flush()
    
    def chooseType(arg):
        if arg.kwargs.get('action') == 'store_true' or arg.kwargs.get('action') == 'store_const':
            return {'type':'check'}
        if arg.kwargs.get('type') == int:
            return {'type':'spin', 'range': 4096}
        if arg.kwargs.get('choices'):
            return {'type':'combo', 'sub':'ticks', 'list':[val for val in arg.kwargs.get('choices')]}
        if arg.args[0] == '--plug':
            return {'type':'combo', 'list': [''] + importPlugs}
        if arg.args[0] == '--option':
            return {'type':'option'}
        if arg.kwargs.get('type') == float:
            return {'type':'entry', 'sub':'numeric'}
        if arg.kwargs.get('type') == 'credits':
            return {'type':'entry', 'sub':'credits'}
        return {'type':'entry'}
    
    def makeWidget(name, arg, sticky = 'ew', label = True, **kwargs):
        kwargs['sticky'] = sticky
        kwargs['label'] = label
        kwargs['change'] = updArgs
        kwargs['tooltip'] = arg['help']
        if arg == allArgs.get(name):
            kwargs['colspan'] = 1
        else:
            kwargs['colspan'] = 9
        
        widget = arg['widget']
        # print(name + ': ' + str(widget))
        if widget['type'] == 'button':
            kwargs.pop('change')
            kwargs.pop('label')
            kwargs.pop('colspan')
            win.button(name, widget['func'], **kwargs)
        elif widget['type'] == 'check':
            win.check(name, argVals[name] or arg.get('default'), text = name, **kwargs)
        elif widget['type'] == 'spin':
            kwargs['item'] = argVals[name] or arg.get('default') or 0
            win.spin(name, 0 - widget['range'], endValue = widget['range'], **kwargs)
        elif widget['type'] == 'combo':
            kwargs['sticky'] = 'w'
            if widget.get('sub'):
                kwargs['kind'] = widget['sub']
                kwargs.pop('label')
            win.combo(name, widget['list'], **kwargs)
            
            # If we're switching to the 'station' command from another command,
            # this argument needs to be cleared from allArgs to not mess things up.
            if not widget.get('sub'):
                if argVals[name] != str:
                    argVals[name] = None
                default = '?' if arg.get('choices') else ''
                win.combo(name, argVals[name] or arg.get('default') or default, callFunction = False)
            else:
            # If we're switching to another command from the 'station' command,
            # this argument needs to be cleared from allArgs to not mess things up.
                if argVals[name] == str:
                    argVals[name] = None
                if argVals[name]:
                    for val in argVals[name]:
                        win.setOptionBox(name, val, value = argVals[name][val], callFunction = False)
        
        elif widget['type'] == 'option':
            kwargs.pop('change')
            kwargs.pop('label')
            kwargs.pop('colspan')
            win.button('optionButton', optionsWin, name = '--option', **kwargs)
            kwargs['sticky'] = sticky
            kwargs['change'] = updArgs
            win.entry(name, argVals[name] or arg.get('default'), row = 'p' , column = 1, colspan = 9, **kwargs)
        elif widget['type'] == 'entry':
            if widget.get('sub'):
                if widget.get('sub') == 'credits':
                    # TODO: Handle 'credits' type.
                    pass
                else:
                    kwargs['kind'] = 'numeric'
            
            win.entry(name, argVals[name] or arg.get('default'), **kwargs)
    
    def setOpts():
        sw = win.widgetManager.get(WIDGET_NAMES.SubWindow, "Plugin Options")
        plugOpts = allArgs['import']['opt']['--option']['options'].get(win.combo('--plug'))
        argStr = ''
        # print(plugOpts)
        if plugOpts:
            for option in plugOpts:
                # print(option + ': ' + plugOpts[option])
                if '=' in plugOpts[option]:
                    if win.entry(option):
                        argStr = argStr + option + '=' + win.entry(option) + ','
                elif win.check(option):
                    argStr = argStr + option + ','
            # print(argStr)
            argStr = argStr.rsplit(',', 1)[0]
            win.entry('--option', argStr)
        sw.hide()
    
    def optionsWin():
        with win.subWindow("Plugin Options", modal = True) as sw:
            win.emptyCurrentContainer()
            optDict = {}
            if argVals['--option']:
                for option in enumerate(argVals['--option'].split(',')):
                    if '=' in option[1]:
                        optDict[option[1].split('=')[0]] = option[1].split('=')[1]
                    else:
                        if option[1] != '':
                            optDict[option[1]] = True
            # print(optDict)
            if not win.combo('--plug'):
                win.message('No import plugin chosen.', width = 170, colspan = 10)
            else:
                plugOpts = allArgs['import']['opt']['--option']['options'][win.combo('--plug')]
                for option in plugOpts:
                    # print(option + ': ' + plugOpts[option])
                    if '=' in plugOpts[option]:
                        win.entry(option, optDict.get(option) or '', label = True, sticky = 'ew', colspan = 10, tooltip = plugOpts[option])
                    else:
                        win.check(option, optDict.get(option) or False, sticky = 'ew', colspan = 10, tooltip = plugOpts[option])
                # print(plugOpts)
            win.button("Done", setOpts, column = 8)
            win.button("Cancel", sw.hide, row = 'p', column = 9)
            sw.show()
    
    def updArgs(name):
        
        def getWidgetType(name):
            for widgetType in [WIDGET_NAMES.Entry, WIDGET_NAMES.Button, WIDGET_NAMES.CheckBox,
                               WIDGET_NAMES.RadioButton, WIDGET_NAMES.SpinBox, WIDGET_NAMES.OptionBox]:
                try:
                    win.widgetManager.get(widgetType, name)
                    return WIDGET_NAMES.widgets[widgetType]
                except:
                    pass
            return None
        
        # Update the stored value of the argument that's been changed.
        if name == '--plug' and argVals[name] != win.get(getWidgetType(name), name):
            win.entry('--option', '')
        
        argVals[name] = win.get(getWidgetType(name), name)
        # print('Changed: "' + name + '" [' + str(argVals[name]) + ']')
        
        if allArgs.get(name):
            excluded = allArgs[name].get('excludes')
            argBase = allArgs
        else:
            try:
                excluded = allArgs[win.combo("Command")]['opt'][name].get('excludes')
                argBase = allArgs[win.combo("Command")]['opt']
            except KeyError:
                excluded = None
        
        # Reset any arguments excluded by this one if it's been set.
        
        # print("'" + str(argVals[name]) + "': " + str(argVals[name] == True)
        #       + "': " + str(not argVals[name]) + " : " + str(not not argVals[name])
        #      )
        
        # Apparently, with strings, 'not not ""' != '""' when considered as a boolean.
        if excluded and not not argVals[name]:
            for exclude in excluded:
                widgetType = argBase[exclude]['widget']['type']
                if widgetType == 'combo':
                    win.combo(exclude, mode = 'clear', callFunction = False)
                elif widgetType == 'spin':
                    win.spin(exclude, mode = 'clear', callFunction = False)
                elif widgetType == 'entry':
                    win.setEntry(exclude, '', callFunction = False)
                elif widgetType == 'check':
                    win.check(exclude, False, callFunction = False)
    
    def updCmd():
        cmd = win.combo("Command")
        # print(cmd)
        
        win.message("helpText", cmdHelp[cmd])
        win.emptyScrollPane('req')
        win.emptyScrollPane('opt')
        if cmd == 'help':
            return
        if allArgs[cmd]['req']:
            with win.scrollPane('req', disabled = 'horizontal'):
                win.label('Required:', sticky = 'w')
                for key in allArgs[cmd]['req']:
                    makeWidget(key, allArgs[cmd]['req'][key])
        
        if allArgs[cmd]['opt']:
            with win.scrollPane('opt', disabled = 'horizontal'):
                win.label('Optional:', sticky = 'w')
                for key in allArgs[cmd]['opt']:
                    makeWidget(key, allArgs[cmd]['opt'][key])
    
    def changeCWD():
        """
        Opens a folder select dialog.
        """
        cwd = win.directoryBox("Select the top-level folder for TD to work in...", dirName = argVals['--cwd'])
        if cwd:
            argVals['--cwd'] = str(Path(cwd))
        win.label('cwd', argVals['--cwd'])
    
    def changeDB():
        """
        Opens a file select dialog.
        """
        db = win.openBox("Select the TD database file to use...", dirName = str(Path(argVals['--db']).parent),
                          fileTypes = [('Data Base File', '*.db')])
        if db:
            argVals['--db'] = str(Path(db))
        win.label('db', argVals['--db'])
    
    def runTD():
        """
        Executes the TD command selected in the GUI.
        """
        
        from . import tradeexcept
        
        def getVals(arg, argBase):
            curArg = argVals[arg]
            vals = []
            # See string weirdness, above.
            if not not curArg and curArg != argBase[arg].get('default'):
                if arg in ['--detail', '--debug', '--quiet']:
                    pass
                else:
                    vals.append(str(arg))
                widget = argBase[arg]['widget']
                if widget['type'] == 'check':
                    return vals
                if widget.get('sub') == 'ticks':
                    choices = ''
                    for choice in curArg:
                        if curArg[choice]:
                            choices = choices + choice
                    vals.append(choices)
                    return vals
                if widget.get('sub') == 'credits':
                    # TODO: Handle 'credits' type
                    pass
                vals.append(str(curArg))
            if vals == []:
                return None
            return vals
        
        def runTrade():
            # Can't allow the Run button to do anything when a command is running.
            win.button('Run', lambda x : x)
            # Redirect output to the Output tab in the GUI
            outputText = win.widgetManager.get(WIDGET_NAMES.Message, 'outputText')
            oldout = sys.stdout
            sys.stdout = StdoutRedirector(outputText)
            print('TD command: "' + ' '.join(argv) + '"')
            
            try:
                try:
                    try:
                        if "CPROF" in os.environ:
                            import cProfile
                            cProfile.run("trade(argv)")
                        else:
                            trade(argv)
                    except PluginException as e:
                        print("PLUGIN ERROR: {}".format(e))
                        if 'EXCEPTIONS' in os.environ:
                            raise e
                        sys.exit(1)
                    except tradeexcept.TradeException as e:
                        print("%s: %s" % (argv[0], str(e)))
                        if 'EXCEPTIONS' in os.environ:
                            raise e
                        sys.exit(1)
                except (UnicodeEncodeError, UnicodeDecodeError) as e:
                    print("-----------------------------------------------------------")
                    print("ERROR: Unexpected unicode error in the wild!")
                    print()
                    print(traceback.format_exc())
                    print(
                        "Please report this bug (http://kfs.org/td/issues). You may be "
                        "able to work around it by using the '-q' parameter. Windows "
                        "users may be able to use 'chcp.com 65001' to tell the console "
                        "you want to support UTF-8 characters."
                        )
            except SystemExit as e:
                print(e)
            
            print("Execution complete.")
            # Set the output back to normal
            sys.stdout = oldout
            # Allow the Run button to do something again.
            win.button('Run', runTD)
        
        win.setTabbedFrameSelectedTab('tabFrame', 'Output')
        
        cmd = win.combo("Command")
        argv = ['trade', cmd ]
        if cmd != 'help':
            for arg in allArgs[cmd]['req']:
                result = getVals(arg, allArgs[cmd]['req'])
                if result:
                    #not all required args don't include the arg name
                    argv = argv + (result if '-' in result[0] else result[1:])
            
            for arg in allArgs[cmd]['opt']:
                result = getVals(arg, allArgs[cmd]['opt'])
                if result:
                    argv = argv + result
            
            for arg in allArgs:
                if arg in Commands:
                    continue
                result = getVals(arg, allArgs)
                if result:
                    argv = argv + result
        
        win.message('outputText', '')
        threading.Thread(target = runTrade, name = "TDThread", daemon = True).start()
    
    # All available commands
    Commands = ['help'] + [ cmd for cmd, module in sorted(commands.commandIndex.items()) ]
    # Used to run TD cli commands.
    cmdenv = commands.CommandIndex().parse
    # 'help' output, required & optional arguments, for each command
    cmdHelp = {}
    
    dbS = str(Path((os.environ.get('TD_DATA') or os.path.join(os.getcwd(), 'data')) / Path('TradeDangerous.db')))
    cwdS = str(Path(os.getcwd()))
    allArgs = {
        '--debug': { 'help': 'Enable/raise level of diagnostic output.',
                    'default':  0, 'required': False, 'action': 'count',
                    'widget': {'type':'combo', 'list': ['', '-w', '-ww', '-www']}
                    },
        '--detail':{ 'help': 'Increase level of detail in output.',
                    'default': 0, 'required': False, 'action': 'count',
                    'excludes': ['--quiet'], 'widget': {'type':'combo', 'list': ['', '-v', '-vv', '-vvv']}
                    },
        '--quiet':{ 'help': 'Reduce level of detail in output.',
                   'default': 0, 'required': False, 'action': 'count',
                   'excludes': ['--detail'], 'widget': {'type':'combo', 'list': ['', '-q', '-qq', '-qqq']}
                   },
        '--db':{ 'help': 'Specify location of the SQLite database.',
                'default': dbS, 'dest': 'dbFilename', 'type': str,
                'widget': {'type':'button', 'func':changeDB}
                },
        '--cwd':{ 'help': 'Change the working directory file accesses are made from.',
                 'default': cwdS, 'type': str, 'required': False,
                 'widget': {'type':'button', 'func':changeCWD}
                 },
        '--link-ly':{ 'help': 'Maximum lightyears between systems to be considered linked.',
                     'default': 30,
                     'widget': {'type':'entry', 'sub':'numeric'}
                     }
        }
    
    # Used to save the value of the arguments.
    argVals = {'--debug': '',
               '--detail': '',
               '--quiet': '',
               '--db': dbS,
               '--cwd': cwdS,
               '--link-ly': 30
               }
    
    sys.argv = ['trade']
    if not argv:
        argv = sys.argv
    if sys.hexversion < 0x03040200:
        raise SystemExit(
            "Sorry: TradeDangerous requires Python 3.4.2 or higher.\n"
            "For assistance, see:\n"
            "\tBug Tracker: https://github.com/eyeonus/Trade-Dangerous/issues\n"
            "\tDocumentation: https://github.com/eyeonus/Trade-Dangerous/wiki\n"
            "\tEDForum Thread: https://forums.frontier.co.uk/showthread.php/441509\n"
            )
    
    try:
        cmdenv(['help'])
    except exceptions.UsageError as e:
        cmdHelp['help'] = str(e)
    for cmd in Commands:
        # print(cmd)
        if cmd == 'help':
            continue
        try:
            cmdenv(['trade', cmd, '-h'])
        except exceptions.UsageError as e:
            cmdHelp[cmd] = str(e)
        index = commands.commandIndex[cmd]
        
        allArgs[cmd] = {'req': {}, 'opt': {}}
        if index.arguments:
            for arg in index.arguments:
                # print(arg.args[0])
                argVals[arg.args[0]] = arg.kwargs.get('default') or None
                
                allArgs[cmd]['req'][arg.args[0]] = {kwarg : arg.kwargs[kwarg] for kwarg in arg.kwargs}
                allArgs[cmd]['req'][arg.args[0]]['widget'] = chooseType(arg)
        # print(allArgs[cmd]['req'])
        
        if index.switches:
            for arg in index.switches:
                try:
                    argVals[arg.args[0]] = arg.kwargs.get('default') or None
                    
                    allArgs[cmd]['opt'][arg.args[0]] = {kwarg : arg.kwargs[kwarg] for kwarg in arg.kwargs}
                    allArgs[cmd]['opt'][arg.args[0]]['widget'] = chooseType(arg)
                    
                    if arg.args[0] == '--option':
                        # Currently only the 'import' cmd has the '--plug' option,
                        # but this could no longer be the case in future.
                        if cmd == 'import':
                            plugOptions = {
                                plug: plugins.load(cmdenv(['trade', cmd, '--plug', plug, '-O', 'help']).plug,
                                                    "ImportPlugin").pluginOptions for plug in importPlugs
                                }
                            allArgs[cmd]['opt'][arg.args[0]]['options'] = plugOptions
                
                except AttributeError:
                    for argGrp in arg.arguments:
                        argVals[argGrp.args[0]] = argGrp.kwargs.get('default') or None
                        
                        allArgs[cmd]['opt'][argGrp.args[0]] = {kwarg : argGrp.kwargs[kwarg] for kwarg in argGrp.kwargs}
                        allArgs[cmd]['opt'][argGrp.args[0]]['widget'] = chooseType(argGrp)
                        
                        allArgs[cmd]['opt'][argGrp.args[0]]['excludes'] = [excl.args[0] for excl in arg.arguments 
                                                                   if excl.args[0] != argGrp.args[0]]
                        if argGrp.args[0] == '--plug':
                            # Currently only the 'import' cmd has the '--plug' option,
                            # but this could no longer be the case in future.
                            if cmd == 'import':
                                allArgs[cmd]['opt'][argGrp.args[0]]['plugins'] = importPlugs
        # print(allArgs[cmd]['opt'])
    # print(allArgs)
    # print(argVals)
    
    with gui('Trade Dangerous GUI (Beta), TD v.%s' % (__version__,), inPadding = 1) as win:
        win.setFont(size = 8, family = 'Courier')
        win.combo('Command', Commands, change = updCmd, tooltip = 'Trade Dangerous command to run.',
                  stretch = 'none', sticky = 'ew', width = 10, row = 0, column = 0, colspan = 5)
        with win.scrollPane('req', disabled = 'horizontal', row = 1, column = 0, colspan = 10) as pane:
            pane.configure(width = 200, height = 75)
        
        with win.scrollPane('opt', disabled = 'horizontal', row = 2, column = 0, colspan = 10) as pane:
            pane.configure(width = 200, height = 345)
        
        with win.tabbedFrame('tabFrame', disabled = 'horizontal', row = 1, column = 10, rowspan = 2, colspan = 40) as tabFrame:
            with win.tab('Help'):
                with win.scrollPane('helpPane', disabled = 'horizontal') as pane:
                    pane.configure(width = 560, height = 420)
                    win.message('helpText', cmdHelp['help'])
                    win.widgetManager.get(WIDGET_NAMES.Message, 'helpText').config(width = 560)
            
            with win.tab('Output'):
                with win.scrollPane('outPane', disabled = 'horizontal') as pane:
                    pane.configure(width = 560, height = 420)
                    win.message('outputText', '')
                    win.widgetManager.get(WIDGET_NAMES.Message, 'outputText').config(width = 560)
        
        makeWidget('--link-ly', allArgs['--link-ly'], sticky = 'w', width = 4, row = 3, column = 2)
        
        makeWidget('--quiet', allArgs['--quiet'], sticky = 'e', disabled = ':', width = 1, row = 3, column = 46)
        
        makeWidget('--detail', allArgs['--detail'], sticky = 'e', disabled = ':', width = 1, row = 3, column = 47)
        
        makeWidget('--debug', allArgs['--debug'], sticky = 'e', disabled = ':', width = 1, row = 3, column = 48)
        
        win.button('Run', runTD, tooltip = 'Execute the selected command.',
                   sticky = 'w', row = 3, column = 49)
        
        makeWidget('--cwd', allArgs['--cwd'], width = 4, row = 4, column = 0)
        with win.scrollPane('CWD', disabled = 'vertical', row = 4, column = 1, colspan = 49) as pane:
            pane.configure(width = 500, height = 20)
            win.label('cwd', argVals['--cwd'], sticky = 'w')
        
        makeWidget('--db', allArgs['--db'], width = 4, row = 5, column = 0)
        with win.scrollPane('DB', disabled = 'vertical', row = 5, column = 1, colspan = 49) as pane:
            pane.configure(width = 500, height = 20)
            win.label('db', argVals['--db'], sticky = 'w')
    
    # End of window


def trade(argv):
    """
    This method represents the trade command.
    """
    cmdIndex = commands.CommandIndex()
    cmdenv = cmdIndex.parse(argv)
    
    tdb = tradedb.TradeDB(cmdenv, load = cmdenv.wantsTradeDB)
    if cmdenv.usesTradeData:
        tsc = tdb.tradingStationCount
        if tsc == 0:
            raise exceptions.NoDataError(
                "There is no trading data for ANY station in "
                "the local database. Please enter or import "
                "price data."
            )
        if tsc == 1:
            raise exceptions.NoDataError(
                "The local database only contains trading data "
                "for one station. Please enter or import data "
                "for additional stations."
            )
        if tsc < 8:
            cmdenv.NOTE(
                "The local database only contains trading data "
                "for {} stations. Please enter or import data "
                "for additional stations.".format(
                    tsc
                )
            )
    
    try:
        results = cmdenv.run(tdb)
    finally:
        # always close tdb
        tdb.close()
    
    if results:
        results.render()
